# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.


import contextlib
import copy
import re
from urllib.parse import urlparse

from glom import glom, assign


def int_or_none(data):
    try:
        return int(data)
    except (TypeError, ValueError):
        return None


def get_crashing_thread(crash_data):
    """
    Takes crash data and returns the crashing thread as an int.

    :arg crash_data: crash data structure that conforms to the schema

    :returns: crashing thread as an int

    """
    try:
        return int(crash_data.get("crashing_thread", 0))
    except (TypeError, ValueError):
        return 0


DOES_NOT_EXIST = object()


@contextlib.contextmanager
def override_values(crash_data, values):
    """
    Takes a dict of path -> value to override in the original crash data.
    After the context is over, the crash data will return to the original
    value.

    :arg crash_data: the crash data that conforms to the schema
    :arg values: dict of path -> value to override

    :yields: dict with overridden values

    """
    crash_data = copy.deepcopy(crash_data)

    for path, value in values.items():
        assign(crash_data, path, val=value, missing=dict)

    yield crash_data


def convert_to_crash_data(processed_crash):
    """
    Takes a processed crash (these are Socorro-centric data structures) and converts
    them to a crash data structure used by signature generation.

    :arg processed_crash: processed crash data from Socorro

    :returns: crash data structure that conforms to the schema

    """
    crash_data = {
        # text or None
        "hang": glom(processed_crash, "hang", default=None),
        # int or None
        "oom_allocation_size": int_or_none(
            glom(processed_crash, "oom_allocation_size", default=None)
        ),
        # str or None
        "js_large_allocation_failure": glom(
            processed_crash,
            "js_large_allocation_failure",
            default=None,
        ),
        # text or None
        "abort_message": glom(processed_crash, "abort_message", default=None),
        # text or None
        "mdsw_status_string": glom(processed_crash, "mdsw_status_string", default=None),
        # text json with "phase", "conditions" (complicated--see code) or None
        "async_shutdown_timeout": glom(
            processed_crash, "async_shutdown_timeout", default=None
        ),
        # text or None
        "ipc_channel_error": glom(processed_crash, "ipc_channel_error", default=None),
        # text or None
        "ipc_message_name": glom(processed_crash, "ipc_message_name", default=None),
        # text
        "moz_crash_reason": glom(processed_crash, "moz_crash_reason", default=None),
        # list of str; for example:
        # ["upload_file_minidump_browser", "upload_file_minidump_content"]
        "additional_minidumps": glom(
            processed_crash, "additional_minidumps", default=[]
        ),
        # pull out the original signature if there was one
        "original_signature": glom(processed_crash, "signature", default=""),
    }

    # Add platform-specific data for stacks

    if "json_dump" in processed_crash:
        # int or None
        crash_data["crashing_thread"] = glom(
            processed_crash, "json_dump.crash_info.crashing_thread", default=None
        )
        # string or None
        crash_data["reason"] = glom(
            processed_crash, "json_dump.crash_info.type", default=None
        )
        # list of CStackTrace or None
        crash_data["threads"] = glom(processed_crash, "json_dump.threads", default=None)
        # text or None
        crash_data["os"] = glom(
            processed_crash, "json_dump.system_info.os", default=None
        )

    if "java_exception" in processed_crash:
        # java_exception structure or {}
        crash_data["java_exception"] = glom(processed_crash, "java_exception")

    if "java_stack_trace" in processed_crash:
        # java_stack_trace string or None
        crash_data["java_stack_trace"] = glom(processed_crash, "java_stack_trace")

    return crash_data


#: List of allowed characters: ascii, printable, and non-whitespace except space
ALLOWED_CHARS = [chr(c) for c in range(32, 127)]


def drop_bad_characters(text):
    """Takes a text and drops all non-printable and non-ascii characters and
    also any whitespace characters that aren't space.

    :arg str text: the text to fix

    :returns: text with all bad characters dropped

    """
    # Strip all non-ascii and non-printable characters
    text = "".join([c for c in text if c in ALLOWED_CHARS])
    return text


def parse_source_file(source_file):
    """Parses a source file thing and returns the file name

    Example:

    >>> parse_file('hg:hg.mozilla.org/releases/mozilla-esr52:js/src/jit/MIR.h:755067c14b06')
    'js/src/jit/MIR.h'

    :arg str source_file: the source file ("file") from a stack frame

    :returns: the filename or ``None`` if it couldn't determine one

    """
    if not source_file:
        return None

    vcsinfo = source_file.split(":")
    if len(vcsinfo) == 4:
        # These are repositories or cloud file systems (e.g. hg, git, s3)
        vcstype, root, vcs_source_file, revision = vcsinfo
        return vcs_source_file

    if len(vcsinfo) == 2:
        # These are directories on someone's Windows computer and vcstype is a
        # file system (e.g. "c:", "d:", "f:")
        vcstype, vcs_source_file = vcsinfo
        return vcs_source_file

    if source_file.startswith("/"):
        # These are directories on OSX or Linux
        return source_file

    # We have no idea what this is, so return None
    return None


def _is_exception(exceptions, before_token, after_token, token):
    """Predicate for whether the open token is in an exception context

    :arg exceptions: list of strings or None
    :arg before_token: the text of the function up to the token delimiter
    :arg after_token: the text of the function after the token delimiter
    :arg token: the token (only if we're looking at a close delimiter

    :returns: bool

    """
    if not exceptions:
        return False
    for s in exceptions:
        if before_token.endswith(s):
            return True
        if s in token:
            return True
    return False


def collapse(function, open_string, close_string, replacement="", exceptions=None):
    """Collapses the text between two delimiters in a frame function value

    This collapses the text between two delimiters and either removes the text
    altogether or replaces it with a replacement string.

    There are certain contexts in which we might not want to collapse the text
    between two delimiters. These are denoted as "exceptions" and collapse will
    check for those exception strings occuring before the token to be replaced
    or inside the token to be replaced.

    Before::

        IPC::ParamTraits<nsTSubstring<char> >::Write(IPC::Message *,nsTSubstring<char> const &)
               ^        ^ open token
               exception string occurring before open token

    Inside::

        <rayon_core::job::HeapJob<BODY> as rayon_core::job::Job>::execute
        ^                              ^^^^ exception string inside token
        open token

    :arg function: the function value from a frame to collapse tokens in
    :arg open_string: the open delimiter; e.g. ``(``
    :arg close_string: the close delimiter; e.g. ``)``
    :arg replacement: what to replace the token with; e.g. ``<T>``
    :arg exceptions: list of strings denoting exceptions where we don't want
        to collapse the token

    :returns: new function string with tokens collapsed

    """
    collapsed = []
    open_count = 0
    open_token = []

    for i, char in enumerate(function):
        if not open_count:
            if char == open_string and not _is_exception(
                exceptions, function[:i], function[i + 1 :], ""
            ):
                open_count += 1
                open_token = [char]
            else:
                collapsed.append(char)

        else:
            if char == open_string:
                open_count += 1
                open_token.append(char)

            elif char == close_string:
                open_count -= 1
                open_token.append(char)

                if open_count == 0:
                    token = "".join(open_token)
                    if _is_exception(
                        exceptions, function[:i], function[i + 1 :], token
                    ):
                        collapsed.append("".join(open_token))
                    else:
                        collapsed.append(replacement)
                    open_token = []
            else:
                open_token.append(char)

    if open_count:
        token = "".join(open_token)
        if _is_exception(exceptions, function[:i], function[i + 1 :], token):
            collapsed.append("".join(open_token))
        else:
            collapsed.append(replacement)

    return "".join(collapsed)


def drop_prefix_and_return_type(function):
    """Takes the function value from a frame and drops prefix and return type

    For example::

        static void * Allocator<MozJemallocBase>::malloc(unsigned __int64)
        ^      ^^^^^^ return type
        prefix

    This gets changes to this::

        Allocator<MozJemallocBase>::malloc(unsigned __int64)

    This tokenizes on space, but takes into account types, generics, traits,
    function arguments, and other parts of the function signature delimited by
    things like `', <>, {}, [], and () for both C/C++ and Rust.

    After tokenizing, this returns the last token since that's comprised of the
    function name and its arguments.

    :arg function: the function value in a frame to drop bits from

    :returns: adjusted function value

    """
    DELIMITERS = {"(": ")", "{": "}", "[": "]", "<": ">", "`": "'"}
    OPEN = DELIMITERS.keys()
    CLOSE = DELIMITERS.values()

    # The list of tokens accumulated so far
    tokens = []

    # Keeps track of open delimiters so we can match and close them
    levels = []

    # The current token we're building
    current = []

    for _, char in enumerate(function):
        if char in OPEN:
            levels.append(char)
            current.append(char)
        elif char in CLOSE:
            if levels and DELIMITERS[levels[-1]] == char:
                levels.pop()
                current.append(char)
            else:
                # This is an unmatched close.
                current.append(char)
        elif levels:
            current.append(char)
        elif char == " ":
            tokens.append("".join(current))
            current = []
        else:
            current.append(char)

    if current:
        tokens.append("".join(current))

    while len(tokens) > 1 and tokens[-1].startswith(("(", "[clone")):
        # It's possible for the function signature to have a space between
        # the function name and the parenthesized arguments or [clone ...]
        # thing. If that's the case, we join the last two tokens. We keep doing
        # that until the last token is nice.
        #
        # Example:
        #
        #     somefunc (int arg1, int arg2)
        #             ^
        #     somefunc(int arg1, int arg2) [clone .cold.111]
        #                                 ^
        #     somefunc(int arg1, int arg2) [clone .cold.111] [clone .cold.222]
        #                                 ^                 ^
        tokens = tokens[:-2] + [" ".join(tokens[-2:])]

    return tokens[-1]


CRASH_ID_RE = re.compile(
    r"""
    ^
    [a-f0-9]{8}-
    [a-f0-9]{4}-
    [a-f0-9]{4}-
    [a-f0-9]{4}-
    [a-f0-9]{6}
    [0-9]{6}      # date in YYMMDD
    $
""",
    re.VERBOSE,
)


def is_crash_id_valid(crash_id):
    """Returns whether this is a valid crash id

    :arg str crash_id: the crash id in question

    :returns: True if it's valid, False if not

    """
    return bool(CRASH_ID_RE.match(crash_id))


def parse_crashid(item):
    """Returns a crashid from a number of formats.

    This handles the following three forms of crashids:

    * CRASHID
    * bp-CRASHID
    * http[s]://HOST[:PORT]/report/index/CRASHID

    :arg str item: the thing to parse a crash id from

    :returns: crashid as str or None

    """
    if is_crash_id_valid(item):
        return item

    if item.startswith("bp-") and is_crash_id_valid(item[3:]):
        return item[3:]

    if item.startswith("http"):
        parsed = urlparse(item)
        path = parsed.path
        if path.startswith("/report/index"):
            crash_id = path.split("/")[-1]
            if is_crash_id_valid(crash_id):
                return crash_id


def strip_leading_zeros(text):
    """Strips leading zeros from a hex string.

    Example:

    >>> strip_leading_zeros("0x0000000000032ec0")
    "0x32ec0"

    :param text: the text to strip leading zeros from

    :returns: stripped text

    """
    try:
        return hex(int(text, base=16))
    except (ValueError, TypeError):
        return text



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import importlib
from unittest import mock


# NOTE(willkg): We do this so that we can extract signature generation into its
# own namespace as an external library. This allows the tests to run if it's in
# "siggen" or "socorro.signature".
base_module = ".".join(__name__.split(".")[:-2])
generator = importlib.import_module(base_module + ".generator")


class TestSignatureGenerator:
    def test_empty_dicts(self):
        generator_obj = generator.SignatureGenerator()
        ret = generator_obj.generate({})

        # NOTE(willkg): This is what the current pipeline yields. If any of those parts change, this
        # might change, too. The point of this test is that we can pass in empty dicts and the
        # SignatureGenerator and the rules in the default pipeline don't fall over.
        assert ret.signature == "EMPTY: no frame data available"
        assert ret.notes == [
            "SignatureGenerationRule: CSignatureTool: no frame data for crashing thread (0)"
        ]

    def test_failing_rule(self):
        class BadRule:
            pass

        generator_obj = generator.SignatureGenerator(ruleset=[BadRule])
        ret = generator_obj.generate({})

        assert ret.signature == ""
        assert ret.notes == [
            "BadRule: Rule failed: 'BadRule' object has no attribute 'predicate'"
        ]

    def test_error_handler(self):
        exc_value = Exception("Cough")

        class BadRule:
            def predicate(self, crash_data, result):
                raise exc_value

        error_handler = mock.MagicMock()

        generator_obj = generator.SignatureGenerator(
            ruleset=[BadRule], error_handler=error_handler
        )
        generator_obj.generate({"uuid": "ou812"})

        # Make sure error_handler was called with right extra
        assert error_handler.call_args_list == [
            mock.call(
                {"uuid": "ou812"},
                exc_info=(Exception, exc_value, mock.ANY),
                extra={"rule": "BadRule"},
            )
        ]



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from contextlib import suppress
from functools import partial
from itertools import islice
import json
import re

from glom import glom

from .siglists_utils import get_signature_list_content
from .utils import (
    collapse,
    drop_bad_characters,
    drop_prefix_and_return_type,
    get_crashing_thread,
    override_values,
    parse_source_file,
    strip_leading_zeros,
)


SIGNATURE_MAX_LENGTH = 255
MAXIMUM_FRAMES_TO_CONSIDER = 40


def join_ignore_empty(delimiter, list_of_strings):
    return delimiter.join(x for x in list_of_strings if x)


class Rule:
    """Base class for Signature generation rules"""

    @property
    def name(self):
        # NOTE(willkg): In order for this to work in siggen and in Socorro, we can't
        # include the Python module path
        return self.__class__.__name__

    def predicate(self, crash_data, result):
        """Whether or not to run this rule

        :arg dict crash_data: the data to use to generate the signature
        :arg dict result: the current signature generation result

        :returns: True or False

        """
        return True

    def action(self, crash_data, result):
        """Runs the rule against the data

        .. Note::

           This modifies ``result`` in place.

        :arg dict crash_data: the data to use to generate the signature
        :arg dict result: the current signature generation result

        :returns: True

        """
        return True


class CSignatureTool:
    """Generates signature from C/C++/Rust stacks.

    This is the class for signature generation tools that work on C/C++/Rust stacks. It
    normalizes frames and then runs them through the siglists to determine which frames
    should be part of the signature.

    """

    hang_prefixes = {-1: "hang", 1: "chromehang"}

    def __init__(self, datadir=None):
        """
        :param datadir: the directory holding signature lists to use or ``None`` to
            use included signature lists
        """
        super().__init__()

        self.datadir = datadir
        if datadir is not None:
            get_contents = partial(get_signature_list_content, source=datadir)
        else:
            get_contents = get_signature_list_content

        self.irrelevant_signature_re = self.build_re(
            get_contents("irrelevant_signature_re")
        )
        self.prefix_signature_re = self.build_re(get_contents("prefix_signature_re"))
        self.signatures_with_line_numbers_re = self.build_re(
            get_contents("signatures_with_line_numbers_re")
        )
        self.signature_sentinels = get_contents("signature_sentinels")

        self.collapse_arguments = True

        self.fixup_space = re.compile(r" (?=[\*&,])")
        self.fixup_comma = re.compile(r",(?! )")
        self.fixup_hash = re.compile(r"::h[0-9a-fA-F]+$")
        self.fixup_lambda_numbers = re.compile(r"::\$_\d+::")

    def build_re(self, lines):
        return re.compile("|".join(lines))

    def normalize_rust_function(self, function, line):
        """Normalizes a single Rust frame with a function."""
        # Drop the prefix and return type if there is any
        function = drop_prefix_and_return_type(function)

        # Collapse types
        function = collapse(
            function,
            open_string="<",
            close_string=">",
            replacement="<T>",
            exceptions=(" as ",),
        )

        # Collapse arguments
        if self.collapse_arguments:
            function = collapse(
                function, open_string="(", close_string=")", replacement=""
            )

        if self.signatures_with_line_numbers_re.match(function):
            function = f"{function}:{line}"

        # Remove spaces before all stars, ampersands, and commas
        function = self.fixup_space.sub("", function)

        # Ensure a space after commas
        function = self.fixup_comma.sub(", ", function)

        # Remove rust-generated uniqueness hashes
        function = self.fixup_hash.sub("", function)

        return function

    def normalize_cpp_function(self, function, line):
        """Normalizes a single cpp frame with a function"""
        # Drop member function cv/ref qualifiers like const, const&, const &,
        # &, and &&
        for ref in ("const", "const&", "const &", "&&", "&"):
            if function.endswith(ref):
                function = function[: -len(ref)].strip()

        # Convert `anonymous namespace' to (anonymous namespace)

        # Drop the prefix and return type if there is any if it's not operator
        # overloading--operator overloading syntax doesn't have the things
        # we're dropping here and can look curious, so don't try
        if "::operator" not in function:
            function = drop_prefix_and_return_type(function)

        # Normalize `anonymous namespace' to (anonymous namespace). bug #1672847
        function = function.replace("`anonymous namespace'", "(anonymous namespace)")

        # Remove lambda number from frames. bug #1688249
        function = self.fixup_lambda_numbers.sub("::$::", function)

        # Collapse types
        #
        # NOTE(willkg): The " in " is for handling "<unknown in foobar.dll>". bug
        # #1685178
        function = collapse(
            function,
            open_string="<",
            close_string=">",
            replacement="<T>",
            exceptions=("name omitted", "IPC::ParamTraits", " in "),
        )

        # Collapse arguments
        if self.collapse_arguments:
            function = collapse(
                function,
                open_string="(",
                close_string=")",
                replacement="",
                exceptions=("anonymous namespace", "operator"),
            )

        # Remove PGO cold block labels like "[clone .cold.222]". bug #1397926
        if "clone .cold" in function:
            function = collapse(
                function, open_string="[", close_string="]", replacement=""
            )

        if self.signatures_with_line_numbers_re.match(function):
            function = f"{function}:{line}"

        # Remove spaces before all stars, ampersands, and commas
        function = self.fixup_space.sub("", function)

        # Ensure a space after commas
        function = self.fixup_comma.sub(", ", function)

        return function

    def normalize_frame(
        self,
        module=None,
        function=None,
        file=None,
        line=None,
        module_offset=None,
        offset=None,
        unloaded_modules=None,
    ):
        """Normalizes a single frame

        Returns a structured conglomeration of the input parameters to serve as a
        signature. The parameter names of this function reflect the exact names of the
        fields from the JSON MDSW frame output. This allows this function to be invoked
        by passing a frame as ``**a_frame``.

        """
        if function:
            # If there's a filename and it ends in .rs, then normalize using
            # Rust rules
            if file and (parse_source_file(file) or "").endswith(".rs"):
                return self.normalize_rust_function(function=function, line=line)

            # Otherwise normalize it with C/C++ rules
            return self.normalize_cpp_function(function=function, line=line)

        # If there's a file and line number, use that
        if file and line:
            filename = file.rstrip("/\\")
            if "\\" in filename:
                file = filename.rsplit("\\")[-1]
            else:
                file = filename.rsplit("/")[-1]
            return f"{file}#{line}"

        # If there's an offset and no module/module_offset, use that
        if not module and not module_offset:
            if unloaded_modules:
                # Use the first unloaded module and the offset or "0"
                unloaded_module = unloaded_modules[0]
                unloaded_module_module = unloaded_module.get("module")
                unloaded_module_offsets = unloaded_module.get("offsets")
                if unloaded_module_module and unloaded_module_offsets:
                    return "(unloaded {}@{})".format(
                        unloaded_module_module,
                        strip_leading_zeros(unloaded_module_offsets[0]),
                    )
                elif unloaded_module_module:
                    return "(unloaded {})".format(unloaded_module_module)

            if offset:
                return f"@{strip_leading_zeros(offset)}"

        # Return module/module_offset
        return "{}@{}".format(module or "", strip_leading_zeros(module_offset))

    def frame_generator(self, frames):
        """Yields frames one at a time from a list of frames

        Note: this treats inlines as individual frames and yields those when they're
        available.

        :arg frames: list of frame dicts

        :returns: generator of frame data for normalization

        """
        for frame in frames:
            inlines = frame.get("inlines") or []
            for inline in inlines:
                yield {
                    "module": frame.get("module"),
                    "function": inline.get("function"),
                    "file": inline.get("file"),
                    "line": inline.get("line"),
                    "module_offset": frame.get("module_offset"),
                    "offset": frame.get("offset"),
                }

            yield {
                "module": frame.get("module"),
                "function": frame.get("function"),
                "file": frame.get("file"),
                "line": frame.get("line"),
                "module_offset": frame.get("module_offset"),
                "offset": frame.get("offset"),
                # NOTE(gsvelto): unloaded modules can only appear in non-inlined frames
                "unloaded_modules": frame.get("unloaded_modules"),
            }

    def create_frame_list(self, thread_data, make_modules_lower_case=False):
        """Takes thread data and builds a list of frames

        Note: this treats inlines as individual frames.

        :arg thread_data: dict of thread data
        :arg make_modules_lower_case: whether or not to lowercase module

        :returns: list of normalized frames

        """
        normalized_frames = []
        frames = thread_data.get("frames", [])
        for frame in islice(self.frame_generator(frames), MAXIMUM_FRAMES_TO_CONSIDER):
            # Bug #1544246. In Rust 1.34, the panic symbols are missing the module in
            # symbols files. This fixes that by adding the module.
            frame = fix_missing_module(frame)

            if make_modules_lower_case and frame.get("module"):
                frame["module"] = frame["module"].lower()

            normalized_frame = self.normalize_frame(**frame)
            normalized_frames.append(normalized_frame)
        return normalized_frames

    def generate(self, source_list, crashed_thread=None, delimiter=" | "):
        """Iterate over frames in the crash stack and generate a signature.

        First, we look for a sentinel frame and if we find one, we start with that.
        Otherwise we start at the beginning.

        Then each frame in the stack is handled like this:

        * a prefix of a relevant frame: Append this element to the signature
        * a relevant frame: Append this element and stop looking
        * irrelevant: Append this element only after seeing a prefix frame

        The signature is a ' | ' separated string of normalized frame names.

        """
        notes = []
        debug_notes = []

        if self.datadir is None:
            debug_notes.append("using included signature lists")
        else:
            debug_notes.append(f"using signature lists from {self.datadir}")

        # Shorten source_list to the first sentinel found
        sentinel_locations = []
        for a_sentinel in self.signature_sentinels:
            if type(a_sentinel) == tuple:
                a_sentinel, condition_fn = a_sentinel
                if not condition_fn(source_list):
                    continue
            with suppress(ValueError):
                sentinel_locations.append(source_list.index(a_sentinel))

        if sentinel_locations:
            min_index = min(sentinel_locations)
            debug_notes.append(
                'sentinel; starting at "{}" index {}'.format(
                    source_list[min_index], min_index
                )
            )
            source_list = source_list[min_index:]

        # Get all the relevant frame signatures. Note that these function signatures
        # have already been normalized at this point.
        new_signature_list = []
        for a_signature in source_list:
            # If the signature matches the irrelevant signatures regex, skip to the next frame.
            if self.irrelevant_signature_re.match(a_signature):
                debug_notes.append(f'irrelevant; ignoring: "{a_signature}"')
                continue

            # If the frame signature is a dll, remove the @xxxxx part.
            if (
                not a_signature.startswith("(unloaded")
                and ".dll" in a_signature.lower()
            ):
                a_signature = a_signature.split("@")[0]

                # If this trimmed DLL signature is the same as the previous frame's, skip it.
                if new_signature_list and a_signature == new_signature_list[-1]:
                    continue

            new_signature_list.append(a_signature)

            # If the signature does not match the prefix signatures regex, then it is the last
            # one we add to the list.
            if not self.prefix_signature_re.match(a_signature):
                debug_notes.append(f'not a prefix; stop: "{a_signature}"')
                break

            debug_notes.append(f'prefix; continue iterating: "{a_signature}"')

        signature = delimiter.join(new_signature_list)

        # Handle empty signatures to explain why we failed generating them.
        if signature == "" or signature is None:
            if crashed_thread is None:
                notes.append("CSignatureTool: no crashing thread identified")
                signature = "EMPTY: no crashing thread identified"

            else:
                notes.append(
                    f"CSignatureTool: no frame data for crashing thread ({crashed_thread})"
                )
                if source_list:
                    # The frames were probably all irrelevant, so pick the first one.
                    signature = source_list[0]
                else:
                    # There wasn't any frame data to look at.
                    signature = "EMPTY: no frame data available"

        return signature, notes, debug_notes


class JavaSignatureTool:
    """Signature generation class for signatures based on JavaStackTrace."""

    # The max length of a java exception description--if it's longer than this,
    # drop it
    DESCRIPTION_MAX_LENGTH = 255

    java_line_number_killer = re.compile(r"\.java\:\d+\)$")
    java_hex_addr_killer = re.compile(r"@[0-9a-f]{8}")

    def generate(self, source, delimiter=": "):
        if not isinstance(source, str):
            return (
                "EMPTY: Java stack trace not in expected format",
                ["JavaSignatureTool: stack trace not in expected format"],
                [],
            )

        source_list = [x.strip() for x in source.splitlines()]
        if not source_list:
            return (
                "EMPTY: Java stack trace not in expected format",
                ["JavaSignatureTool: stack trace not in expected format"],
                [],
            )

        notes = []
        debug_notes = []

        try:
            java_exception_class, description = source_list[0].split(":", 1)
            java_exception_class = java_exception_class.strip()

            description = self.java_hex_addr_killer.sub("@<addr>", description)
            description = description.strip()

        except ValueError:
            # It throws a ValueError if the first line doesn't have a ":"
            java_exception_class = source_list[0]
            description = ""
            notes.append(
                "JavaSignatureTool: stack trace line 1 is not in the expected format"
            )

        try:
            java_method = self.java_line_number_killer.sub(".java)", source_list[1])
            if not java_method:
                notes.append("JavaSignatureTool: stack trace line 2 is empty")
        except IndexError:
            notes.append("JavaSignatureTool: stack trace line 2 is missing")
            java_method = ""

        # An error in an earlier version of this code resulted in the colon
        # being left out of the division between the description and the
        # java_method if the description didn't end with "<addr>". This code
        # perpetuates that error while correcting the "<addr>" placement when
        # it is not at the end of the description. See Bug 865142 for a
        # discussion of the issues.
        if description.endswith("<addr>"):
            # at which time the colon placement error is to be corrected
            # just use the following line as the replacement for this entire
            # if/else block
            signature = join_ignore_empty(
                delimiter, (java_exception_class, description, java_method)
            )
        else:
            description_java_method_phrase = join_ignore_empty(
                " ", (description, java_method)
            )
            signature = join_ignore_empty(
                delimiter, (java_exception_class, description_java_method_phrase)
            )

        if len(signature) > self.DESCRIPTION_MAX_LENGTH:
            signature = delimiter.join((java_exception_class, java_method))
            notes.append(
                "JavaSignatureTool: dropped Java exception description due to length"
            )

        return signature, notes, debug_notes


class JavaExceptionSignatureTool:
    """Signature generation class for signatures based on JavaException."""

    def normalize_frame(self, frame):
        frame_module = frame.get("module", "")
        frame_function = frame.get("function", "")
        if frame_module and frame_function:
            frame_signature = ".".join([frame_module, frame_function])
        elif frame_module:
            frame_signature = frame_module
        elif frame_function:
            frame_signature = frame_function
        else:
            frame_signature = ""

        frame_filename = frame.get("filename", "")

        if frame_signature and frame_filename:
            return f"{frame_signature}({frame_filename})"
        elif frame_signature:
            return frame_signature
        return "unknown"

    def generate(self, source):
        if not isinstance(source, dict):
            return (
                "EMPTY: Java exception not in expected format",
                ["JavaExceptionSignatureTool: value not in expected format"],
                [],
            )

        # This looks at the first frame (the one that raised the exception) of the first
        # exception value (the exception that started a series of cascading exceptions).
        stacktrace = glom(source, "exception.values.0.stacktrace", default=None)
        if not stacktrace:
            return (
                "EMPTY: Java exception not in expected format",
                ["JavaExceptionSignatureTool: value not in expected format"],
                [],
            )

        notes = []
        debug_notes = []

        exc_module = stacktrace.get("module", "")
        exc_type = stacktrace.get("type", "Unknown")
        if exc_module and exc_type:
            exc_class = f"{exc_module}.{exc_type}"
        else:
            exc_class = exc_type

        frames = stacktrace.get("frames") or []
        if frames:
            normalized_frames = [self.normalize_frame(frame) for frame in frames]
            signature = f"{exc_class}: at {normalized_frames[0]}"
        else:
            signature = f"{exc_class}"

        return signature, notes, debug_notes


# Map of (file, function) -> fixed function for Rust 1.34 symbols that are
# missing module.
FILE_FUNCTION_TO_FUNCTION = {
    (
        "src/liballoc/raw_vec.rs",
        "capacity_overflow",
    ): "alloc::raw_vec::capacity_overflow",
    ("src/libcore/option.rs", "expect_failed"): "core::option::expect_failed",
    (
        "src/libcore/panicking.rs",
        "panic_bounds_check",
    ): "core::panicking::panic_bounds_check",
    ("src/libcore/panicking.rs", "panic_fmt"): "core::panicking::panic_fmt",
    ("src/libcore/panicking.rs", "panic"): "core::panicking::panic",
    (
        "src/libcore/slice/mod.rs",
        "slice_index_order_fail",
    ): "core::slice::slice_index_order_fail",
    ("src/libstd/panicking.rs", "begin_panic_fmt"): "std::panicking::begin_panic_fmt",
    (
        "src/libstd/panicking.rs",
        "continue_panic_fmt",
    ): "std::panicking::continue_panic_fmt",
    (
        "src/libstd/panicking.rs",
        "rust_panic_with_hook",
    ): "std::panicking::rust_panic_with_hook",
}


def fix_missing_module(frame):
    """Fixes the module for panic symbols generated by Rust 1.34.

    For example, this turns "panic" into "core::panicking::panic". This allows
    signature sentinels to work.

    See bug #1544246

    """
    if "file" not in frame or "function" not in frame:
        return frame
    fn = parse_source_file(frame.get("file", ""))
    fixed_function = FILE_FUNCTION_TO_FUNCTION.get((fn, frame["function"]))
    if fixed_function:
        frame["function"] = fixed_function

    return frame


class SignatureGenerationRule(Rule):
    """Generates a signature based on stack frames.

    For Java crashes, this generates a basic signature using stack frames.

    For C/C++/Rust crashes, this generates a more robust signature using
    normalized versions of stack frames augmented by the contents of the
    signature lists.

    Rough signature list rules (there are more details in the siglists README):

    1. Walk the frames looking for a "signature sentinel" which becomes the
       first item in the signature.
    2. Continue walking frames.

       1. If the frame is in the "irrelevant" list, ignore it and
          continue.
       2. If the frame is in the "prefix" list, add it to the signature
          and continue.
       3. If the frame isn't in either list, stop walking frames.

    3. Signature is generated by joining those frames with " | " between
       them.

    If it's a C/C++/Rust signature, this rule also adds to ``result.extra``:

    * ``normalized_frames``: the list of normalized frames
    * ``proto_signature``: a ``" | "`` delimited string of the normalized
      frames

    :param signature_list_dir: path to the directory with the signature lists to use or
        ``None`` if you want to use the included ones

    """

    def __init__(self, signature_list_dir=None):
        super().__init__()
        self.java_signature_tool = JavaSignatureTool()
        self.java_exception_signature_tool = JavaExceptionSignatureTool()
        self.c_signature_tool = CSignatureTool(datadir=signature_list_dir)

    def action(self, crash_data, result):
        # If this is a Java crash, then generate a Java signature
        if crash_data.get("java_stack_trace"):
            result.debug(self.name, "using JavaSignatureTool")
            signature, notes, debug_notes = self.java_signature_tool.generate(
                crash_data["java_stack_trace"], delimiter=": "
            )
            for note in notes:
                result.info(self.name, note)
            for note in debug_notes:
                result.debug(self.name, note)
            result.set_signature(self.name, signature)
            return True

        # If there's a java_exception, use that to generate the signature
        #
        # NOTE(willkg): At some point, we want to rewrite signature generation for Java
        # crashes and we'll want to make this take precedence over java_stack_trace.
        if crash_data.get("java_exception"):
            result.debug(self.name, "using JavaExceptionSignatureTool")
            signature, notes, debug_notes = self.java_exception_signature_tool.generate(
                crash_data["java_exception"]
            )
            for note in notes:
                result.info(self.name, note)
            for note in debug_notes:
                result.debug(self.name, note)
            result.set_signature(self.name, signature)
            return True

        result.debug(self.name, "using CSignatureTool")
        try:
            crashing_thread = get_crashing_thread(crash_data)

            # If we have a thread to look at, pull the frames for that.
            # Otherwise we don't have frames to use.
            if crashing_thread is not None:
                signature_list = self.c_signature_tool.create_frame_list(
                    glom(crash_data, "threads.%d" % crashing_thread, default={}),
                    crash_data.get("os") == "Windows NT",
                )

            else:
                signature_list = []

        except (KeyError, IndexError) as exc:
            result.note("No crashing frames found because of %s", exc)
            signature_list = []

        signature, notes, debug_notes = self.c_signature_tool.generate(
            signature_list,
            crashing_thread,
        )

        if signature_list:
            # Capture the proto_signature which gets indexed in Elasticsearch
            # and is searchable
            result.extra["proto_signature"] = " | ".join(signature_list)
            # Capture the normalized_frames for users of signature generation
            result.extra["normalized_frames"] = signature_list
        for note in notes:
            result.info(self.name, note)
        for note in debug_notes:
            result.debug(self.name, note)
        if signature:
            result.set_signature(self.name, signature)

        return True


class StackOverflowSignature(Rule):
    """Prepends ``stackoverflow``

    See `bug #1796389 <https://bugzilla.mozilla.org/show_bug.cgi?id=1796389>`__.

    """

    # These reason values indicate a stackoverflow
    stackoverflow_reason = [
        "EXCEPTION_STACK_OVERFLOW",
    ]

    def predicate(self, crash_data, result):
        # Check the reason to see if it's one of a few values that indicate a
        # stackoverflow
        reason = crash_data.get("reason", None)
        if reason in self.stackoverflow_reason:
            return True

        return False

    def action(self, crash_data, result):
        result.set_signature(self.name, f"stackoverflow | {result.signature}")
        return True


class OOMSignature(Rule):
    """Prepends ``OOM | <size>`` to signatures for OOM crashes.

    See `bug #1007530 <https://bugzilla.mozilla.org/show_bug.cgi?id=1007530>`__.

    """

    # Fragments that are in the signature that indicate this is an OOM
    signature_fragments = (
        "NS_ABORT_OOM",
        "mozalloc_handle_oom",
        "CrashAtUnhandlableOOM",
        "AutoEnterOOMUnsafeRegion",
        "alloc::oom::oom",
    )

    # These reason values indicate an OOM
    oom_reason = [
        "STATUS_FATAL_MEMORY_EXHAUSTION",
        "STATUS_NO_MEMORY",
    ]

    # These last_error_value values indicate an OOM
    oom_last_error_value = [
        "ERROR_COMMITMENT_LIMIT",
        "ERROR_NOT_ENOUGH_MEMORY",
    ]

    def predicate(self, crash_data, result):
        if crash_data.get("oom_allocation_size"):
            return True

        signature = result.signature
        if not signature:
            return False

        for a_signature_fragment in self.signature_fragments:
            if a_signature_fragment in signature:
                return True

        # Check the last_error_value of the crashing thread to see if it's
        # ERROR_COMMITMENT_LIMIT which indicates this is an OOM
        crashing_thread = get_crashing_thread(crash_data)
        thread = glom(crash_data, "threads.%d" % crashing_thread, default={})
        if thread:
            last_error_value = thread.get("last_error_value", "")
            if last_error_value in self.oom_last_error_value:
                return True

        # Check the reason to see if it's one of a few values that indicate an OOM
        reason = crash_data.get("reason", None)
        if reason in self.oom_reason:
            return True

        # Check js_large_allocation_failure to see if it's Reporting
        js_large_allocation_failure = crash_data.get(
            "js_large_allocation_failure", None
        )
        if js_large_allocation_failure == "Reporting":
            return True
        return False

    def action(self, crash_data, result):
        if crash_data.get("js_large_allocation_failure") == "Reporting":
            # If a JSLargeAllocationFailure was Reporting, then it's an OOM | large
            result.set_signature(self.name, f"OOM | large | {result.signature}")
            return True

        try:
            size = int(crash_data.get("oom_allocation_size"))
        except (TypeError, AttributeError, KeyError):
            result.set_signature(self.name, f"OOM | unknown | {result.signature}")
            return True

        if size <= 262144:  # 256K
            result.set_signature(self.name, "OOM | small")
        else:
            result.set_signature(self.name, f"OOM | large | {result.signature}")
        return True


class BadHardware(Rule):
    """Prepends ``bad hardware`` to signatures that are from bad hardware.

    See `bug #1733904 <https://bugzilla.mozilla.org/show_bug.cgi?id=1733904>`__.

    """

    # These reason value substrings indicate an error caused by bad hardware
    bad_hardware_reason = [
        "STATUS_DEVICE_DATA_ERROR",
    ]

    def predicate(self, crash_data, result):
        reason = crash_data.get("reason", None)
        if not reason:
            return False

        for possibility in self.bad_hardware_reason:
            if possibility in reason:
                return True

        return False

    def action(self, crash_data, result):
        result.set_signature(self.name, f"bad hardware | {result.signature}")
        return True


class AbortSignature(Rule):
    """Prepends abort message to signature.

    See `bug #803779 <https://bugzilla.mozilla.org/show_bug.cgi?id=803779>`__.

    """

    def predicate(self, crash_data, result):
        return bool(crash_data.get("abort_message"))

    def action(self, crash_data, result):
        abort_message = crash_data["abort_message"]

        if "###!!! ABORT: file " in abort_message:
            # This is an abort message that contains no interesting
            # information. We just want to put the "Abort" marker in the
            # signature.
            result.set_signature(self.name, f"Abort | {result.signature}")
            return True

        if "###!!! ABORT:" in abort_message:
            # Recent crash reports added some irrelevant information at the
            # beginning of the abort message. We want to remove that and keep
            # just the actual abort message.
            abort_message = abort_message.split("###!!! ABORT:", 1)[1]

        if ": file " in abort_message:
            # Abort messages contain a file name and a line number. Since
            # those are very likely to change between builds, we want to
            # remove those parts from the signature.
            abort_message = abort_message.split(": file ", 1)[0]

        if "unable to find a usable font" in abort_message:
            # "unable to find a usable font" messages include a parenthesized localized message. We
            # want to remove that. Bug #1385966
            open_paren = abort_message.find("(")
            if open_paren != -1:
                end_paren = abort_message.rfind(")")
                if end_paren != -1:
                    abort_message = (
                        abort_message[:open_paren] + abort_message[end_paren + 1 :]
                    )

        abort_message = drop_bad_characters(abort_message).strip()

        if len(abort_message) > 80:
            abort_message = abort_message[:77] + "..."

        result.set_signature(self.name, f"Abort | {abort_message} | {result.signature}")
        return True


class SigFixWhitespace(Rule):
    """Fix whitespace in signatures.

    This does the following:

    * trims leading and trailing whitespace
    * converts all non-space whitespace characters to space
    * reduce consecutive spaces to a single space

    """

    WHITESPACE_RE = re.compile(r"\s")
    CONSECUTIVE_WHITESPACE_RE = re.compile(r"\s\s+")

    def action(self, crash_data, result):
        original_sig = result.signature

        # Trim leading and trailing whitespace
        sig = original_sig.strip()

        # Convert all non-space whitespace characters into spaces
        sig = self.WHITESPACE_RE.sub(" ", sig)

        # Reduce consecutive spaces to a single space
        sig = self.CONSECUTIVE_WHITESPACE_RE.sub(" ", sig)

        if sig != original_sig:
            result.set_signature(self.name, sig)
        return True


class SigTruncate(Rule):
    """Truncates signatures down to ``SIGNATURE_MAX_LENGTH`` characters."""

    def predicate(self, crash_data, result):
        return len(result.signature) > SIGNATURE_MAX_LENGTH

    def action(self, crash_data, result):
        max_length = SIGNATURE_MAX_LENGTH - 3
        result.set_signature(self.name, "{}...".format(result.signature[:max_length]))
        result.info(self.name, "SigTrunc: signature truncated due to length")
        return True


class StackwalkerErrorSignatureRule(Rule):
    """Appends minidump-stackwalker error to signature."""

    def predicate(self, crash_data, result):
        return bool(
            result.signature.startswith("EMPTY")
            and crash_data.get("mdsw_status_string")
        )

    def action(self, crash_data, result):
        result.set_signature(
            self.name,
            "{}; {}".format(result.signature, crash_data["mdsw_status_string"]),
        )
        return True


class SignatureRunWatchDog(Rule):
    """Prepends ``shutdownhang`` to signature for shutdown hang crashes."""

    def __init__(self):
        super().__init__()
        self.signature_generation_rule = SignatureGenerationRule()

    def predicate(self, crash_data, result):
        return "RunWatchdog" in result.signature

    def action(self, crash_data, result):
        # For shutdownhang crashes, we need to use thread 0 instead of the crashing
        # thread. The reason is because those crashes happen artificially when thread 0
        # gets stuck. So whatever the crashing thread is, we don't care about it and
        # only want to know what was happening in thread 0 when it got stuck.
        with override_values(crash_data, {"crashing_thread": 0}) as crash_data:
            ret = self.signature_generation_rule.action(crash_data, result)

        result.set_signature(self.name, f"shutdownhang | {result.signature}")
        return ret


class SignatureShutdownTimeout(Rule):
    """Replaces signature with ``async_shutdown_timeout`` message.

    This supports ``AsyncShutdownTimeout`` annotation values with the following
    structure::

        {
            "phase": <str>,
            "conditions": [
                {
                    "name": <str>,
                    ...
                }
            ]
        }


    """

    def predicate(self, crash_data, result):
        return bool(crash_data.get("async_shutdown_timeout"))

    def action(self, crash_data, result):
        parts = ["AsyncShutdownTimeout"]
        try:
            shutdown_data = json.loads(crash_data["async_shutdown_timeout"])
            if isinstance(shutdown_data.get("phase"), str):
                parts.append(shutdown_data["phase"])
            else:
                parts.append("(unknown)")

            conditions = [
                # NOTE(willkg): The AsyncShutdownTimeout notation condition can either
                # be a string that looks like a "name" or a dict with a "name" in it.
                #
                # This handles both variations.
                str(c["name"] if isinstance(c, dict) else c)
                for c in shutdown_data.get("conditions") or []
            ]
            if conditions:
                conditions.sort()
                parts.append(",".join(conditions))
            else:
                parts.append("(none)")
        except (ValueError, KeyError) as exc:
            parts.append("UNKNOWN")
            result.info(self.name, "Error parsing AsyncShutdownTimeout: %s", exc)

        new_sig = " | ".join(parts)
        result.info(
            self.name,
            'Signature replaced with a Shutdown Timeout signature, was: "%s"',
            result.signature,
        )
        result.set_signature(self.name, new_sig)
        return True


class SignatureIPCChannelError(Rule):
    """Stomps on signature with shutdownkill signature

    Either ``IPCError-browser | ShutDownKill`` or ``IPCError-content | ShutDownKill``.

    """

    def predicate(self, crash_data, result):
        return bool(crash_data.get("ipc_channel_error"))

    def action(self, crash_data, result):
        minidumps = crash_data.get("additional_minidumps") or []
        if "upload_file_minidump_browser" in minidumps:
            new_sig = "IPCError-browser | {}"
        else:
            new_sig = "IPCError-content | {}"
        new_sig = new_sig.format(crash_data["ipc_channel_error"][:100])
        result.info(self.name, "IPC Channel Error stomped on signature")
        result.set_signature(self.name, new_sig)
        return True


class SignatureIPCMessageName(Rule):
    """Appends ``ipc_message_name`` value to signature."""

    def predicate(self, crash_data, result):
        return bool(crash_data.get("ipc_message_name"))

    def action(self, crash_data, result):
        result.set_signature(
            self.name,
            "{} | IPC_Message_Name={}".format(
                result.signature, crash_data["ipc_message_name"]
            ),
        )
        return True


class SignatureParentIDNotEqualsChildID(Rule):
    """Stomp on the signature if ``moz_crash_reason`` is ``parentBuildID != childBuildID``.

    In the case where the assertion fails, then the parent buildid and the child buildid are
    different. This causes a lot of strangeness particularly in symbolification, so the signatures
    end up as junk. Instead, we want to bucket all these together so we replace the signature.

    """

    def predicate(self, crash_data, result):
        value = "MOZ_RELEASE_ASSERT(parentBuildID == childBuildID)"
        return crash_data.get("moz_crash_reason") == value

    def action(self, crash_data, result):
        result.info(
            self.name,
            'Signature replaced with MOZ_RELEASE_ASSERT, was: "%s"',
            result.signature,
        )

        # The MozCrashReason lists the assertion that failed, so we put "!=" in the signature
        result.set_signature(self.name, "parentBuildID != childBuildID")
        return True


class HungProcess(Rule):
    """Prepends ``hang: <hang_type>`` to the signature of hangs.

    See bug `#1826703 <https://bugzilla.mozilla.org/show_bug.cgi?id=1826703>`__.

    """

    def predicate(self, crash_data, result):
        hang_type = crash_data.get("hang", None)
        return hang_type is not None

    def action(self, crash_data, result):
        hang_type = crash_data.get("hang")
        result.set_signature(self.name, f"hang: {hang_type} | {result.signature}")
        return True



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import copy

import pytest

from ..utils import (
    collapse,
    drop_bad_characters,
    drop_prefix_and_return_type,
    get_crashing_thread,
    int_or_none,
    override_values,
    parse_crashid,
    parse_source_file,
    strip_leading_zeros,
)


@pytest.mark.parametrize(
    "data, expected",
    [
        (None, None),
        ("", None),
        ("5", 5),
    ],
)
def test_int_or_none(data, expected):
    assert int_or_none(data) == expected


@pytest.mark.parametrize(
    "crash_data, expected",
    [
        ({}, 0),
        ({"crashing_thread": ""}, 0),
        ({"crashing_thread": "5"}, 5),
    ],
)
def test_get_crashing_thread(crash_data, expected):
    assert get_crashing_thread(crash_data) == expected


@pytest.mark.parametrize(
    "crash_data, overrides, expected",
    [
        ({}, {}, {}),
        ({"a": 1}, {}, {"a": 1}),
        # Try overriding a one-step path
        ({"a": 1, "b": 1}, {"a": 2}, {"a": 2, "b": 1}),
        # Override a path that doesn't exist
        ({}, {"a": 1}, {"a": 1}),
        # Override a multi-step path
        ({"a": {}, "b": 1}, {"a.b.c": 2}, {"a": {"b": {"c": 2}}, "b": 1}),
    ],
)
def test_override_values(crash_data, overrides, expected):
    original_crash_data = copy.deepcopy(crash_data)
    with override_values(crash_data, overrides) as adjusted_crash_data:
        assert adjusted_crash_data == expected

    # Make sure overriding a value doesn't mess up the original structure
    assert crash_data == original_crash_data


@pytest.mark.parametrize(
    "text, expected",
    [
        ("", ""),
        ("", ""),
        ("123", "123"),
        ("123", "123"),
        # Drop non-ascii characters
        ("1\xc6\x8a23", "123"),
        ("1\u018a23", "123"),
        # Drop non-space whitespace characters
        ("\r\n\t1 23", "1 23"),
        ("\r\n\t1 23", "1 23"),
        # Drop non-printable characters
        ("\0\b1 23", "1 23"),
        ("\0\b1 23", "1 23"),
    ],
)
def test_drop_bad_characters(text, expected):
    assert drop_bad_characters(text) == expected


@pytest.mark.parametrize(
    "source_file, expected",
    [
        (
            "hg:hg.mozilla.org/releases/mozilla-release:js/src/vm/JSFunction.cpp:7d280b7e277b82ef282325fefb601c10698e075b",  # noqa
            "js/src/vm/JSFunction.cpp",
        ),
        (
            "git:github.com/rust-lang/rust:src/libcore/cmp.rs:4d90ac38c0b61bb69470b61ea2cccea0df48d9e5",  # noqa
            "src/libcore/cmp.rs",
        ),
        (
            "f:\\dd\\vctools\\crt\\crtw32\\mbstring\\mbsnbico.c",
            "\\dd\\vctools\\crt\\crtw32\\mbstring\\mbsnbico.c",
        ),
        ("d:\\w7rtm\\com\\rpc\\ndrole\\udt.cxx", "\\w7rtm\\com\\rpc\\ndrole\\udt.cxx"),
        (
            "/build/firefox-Kq_6Wg/firefox-54.0+build3/memory/mozjemalloc/jemalloc.c",
            "/build/firefox-Kq_6Wg/firefox-54.0+build3/memory/mozjemalloc/jemalloc.c",
        ),
        (None, None),
    ],
)
def test_parse_source_file(source_file, expected):
    assert parse_source_file(source_file) == expected


@pytest.mark.parametrize(
    "function, expected",
    [
        ("", ""),
        # Test parsing variations
        ("HeapFree", "HeapFree"),
        ("Foo<bar>", "Foo<T>"),
        ("<bar>Foo", "<T>Foo"),
        ("<bar>", "<T>"),
        ("Foo<bar", "Foo<T>"),
        ("Foo<bar <baz> >", "Foo<T>"),
        ("Foo<bar<baz>", "Foo<T>"),
        (
            "CLayeredObjectWithCLS<CCryptoSession>::Release()",
            "CLayeredObjectWithCLS<T>::Release()",
        ),
        (
            "core::ptr::drop_in_place<style::stylist::CascadeData>",
            "core::ptr::drop_in_place<T>",
        ),
        # Test exceptions
        (
            "<rayon_core::job::HeapJob<BODY> as rayon_core::job::Job>::execute",
            "<rayon_core::job::HeapJob<BODY> as rayon_core::job::Job>::execute",
        ),
        ("<name omitted>", "<name omitted>"),
        (
            "IPC::ParamTraits<nsTSubstring<char> >::Write(IPC::Message *,nsTSubstring<char> const &)",
            "IPC::ParamTraits<nsTSubstring<T> >::Write(IPC::Message *,nsTSubstring<T> const &)",
        ),
    ],
)
def test_collapse(function, expected):
    params = {
        "function": function,
        "open_string": "<",
        "close_string": ">",
        "replacement": "<T>",
        "exceptions": ["name omitted", "IPC::ParamTraits", " as "],
    }
    assert collapse(**params) == expected


@pytest.mark.parametrize(
    "function, expected",
    [
        # C/C++
        ("`anonymous namespace'::xClose", "`anonymous namespace'::xClose"),
        (
            "bool CCGraphBuilder::BuildGraph(class js::SliceBudget & const)",
            "CCGraphBuilder::BuildGraph(class js::SliceBudget & const)",
        ),
        (
            "int nsHtml5Tokenizer::stateLoop<nsHtml5SilentPolicy>(int, char16_t, int, char16_t*, bool, int, int)",  # noqa
            "nsHtml5Tokenizer::stateLoop<nsHtml5SilentPolicy>(int, char16_t, int, char16_t*, bool, int, int)",  # noqa
        ),
        (
            "js::ObjectGroup* DoCallback<js::ObjectGroup*>(JS::CallbackTracer*, js::ObjectGroup**, char const*)",  # noqa
            "DoCallback<js::ObjectGroup*>(JS::CallbackTracer*, js::ObjectGroup**, char const*)",
        ),
        (
            "js::Shape* js::Allocate<js::Shape, (js::AllowGC)1>(JSContext*)",
            "js::Allocate<js::Shape, (js::AllowGC)1>(JSContext*)",
        ),
        (
            "long sandbox::TargetNtCreateFile( *, void * *, unsigned long, struct _OBJECT_ATTRIBUTES *, struct _IO_STATUS_BLOCK *, union _LARGE_INTEGER *, unsigned long, unsigned long, unsigned long, unsigned long, void *, unsigned long)",  # noqa
            "sandbox::TargetNtCreateFile( *, void * *, unsigned long, struct _OBJECT_ATTRIBUTES *, struct _IO_STATUS_BLOCK *, union _LARGE_INTEGER *, unsigned long, unsigned long, unsigned long, unsigned long, void *, unsigned long)",  # noqa
        ),
        (
            "static `anonymous-namespace'::reflectStatus `anonymous namespace'::internal_ReflectHistogramAndSamples(struct JSContext *, class JS::Handle<JSObject *>, class base::Histogram *, const class base::Histogram::SampleSet & const)",  # noqa
            "`anonymous namespace'::internal_ReflectHistogramAndSamples(struct JSContext *, class JS::Handle<JSObject *>, class base::Histogram *, const class base::Histogram::SampleSet & const)",  # noqa
        ),
        (
            "static bool `anonymous namespace'::TypeAnalyzer::specializePhis()",
            "`anonymous namespace'::TypeAnalyzer::specializePhis()",
        ),
        (
            "static char * dtoa(struct DtoaState *, union U, int, int, int *, int *, char * *)",
            "dtoa(struct DtoaState *, union U, int, int, int *, int *, char * *)",
        ),
        (
            "static class js::HashSet<js::Shape *,js::ShapeHasher,js::SystemAllocPolicy> * HashChildren(class js::Shape *, class js::Shape *)",  # noqa
            "HashChildren(class js::Shape *, class js::Shape *)",
        ),
        (
            "static class mozilla::dom::Element * GetPropagatedScrollbarStylesForViewport(class nsPresContext *, struct mozilla::ScrollbarStyles *)",  # noqa
            "GetPropagatedScrollbarStylesForViewport(class nsPresContext *, struct mozilla::ScrollbarStyles *)",  # noqa
        ),
        (
            "static const class SkTMaskGamma<3,3,3> & const cached_mask_gamma(float, float, float)",
            "cached_mask_gamma(float, float, float)",
        ),
        (
            "static short ssl_Poll(struct PRFileDesc *, short, short *)",
            "ssl_Poll(struct PRFileDesc *, short, short *)",
        ),
        (
            "static struct already_AddRefed<nsIAsyncShutdownClient> `anonymous namespace'::GetShutdownPhase()",  # noqa
            "`anonymous namespace'::GetShutdownPhase()",
        ),
        (
            "static struct Index * sqlite3FindIndex(struct sqlite3 *, const char *, const char *)",
            "sqlite3FindIndex(struct sqlite3 *, const char *, const char *)",
        ),
        ("static unsigned int pr_root(void *)", "pr_root(void *)"),
        (
            "static void * Allocator<MozJemallocBase>::malloc(unsigned __int64)",
            "Allocator<MozJemallocBase>::malloc(unsigned __int64)",
        ),
        (
            "static void * * NewUNumberFormat(struct JSContext *, class JS::Handle<js::NumberFormatObject *>)",  # noqa
            "NewUNumberFormat(struct JSContext *, class JS::Handle<js::NumberFormatObject *>)",
        ),
        (
            "void mozilla::layers::MLGDeviceD3D11::~MLGDeviceD3D11()",
            "mozilla::layers::MLGDeviceD3D11::~MLGDeviceD3D11()",
        ),
        (
            "void * arena_t::MallocSmall(unsigned int, bool)",
            "arena_t::MallocSmall(unsigned int, bool)",
        ),
        # Rust
        (
            "bool prefs_parser::prefs_parser_parse(char *, prefs_parser::PrefValueKind, char *, unsigned __int64,  *,  *)",  # noqa
            "prefs_parser::prefs_parser_parse(char *, prefs_parser::PrefValueKind, char *, unsigned __int64,  *,  *)",  # noqa
        ),
        (
            "float geckoservo::glue::Servo_AnimationValue_GetOpacity(struct style::gecko_bindings::structs::root::RawServoAnimationValue *)",  # noqa
            "geckoservo::glue::Servo_AnimationValue_GetOpacity(struct style::gecko_bindings::structs::root::RawServoAnimationValue *)",  # noqa
        ),
        (
            "static <NoType> std::panicking::begin_panic<str*>(struct str*, struct (str*, u32, u32) *)",
            "std::panicking::begin_panic<str*>(struct str*, struct (str*, u32, u32) *)",
        ),
        (
            "static core::result::Result style::properties::PropertyDeclaration::to_css(struct nsstring::nsAString *)",  # noqa
            "style::properties::PropertyDeclaration::to_css(struct nsstring::nsAString *)",
        ),
        (
            "static struct atomic_refcell::AtomicRefMut<style::data::ElementData> style::gecko::wrapper::{{impl}}::ensure_data(struct style::gecko::wrapper::GeckoElement *)",  # noqa
            "style::gecko::wrapper::{{impl}}::ensure_data(struct style::gecko::wrapper::GeckoElement *)",
        ),
        (
            "static union core::option::Option<usize> encoding_rs::Encoder::max_buffer_length_from_utf16_without_replacement(unsigned __int64)",  # noqa
            "encoding_rs::Encoder::max_buffer_length_from_utf16_without_replacement(unsigned __int64)",
        ),
        (
            "static unsigned __int64 style::gecko::pseudo_element::PseudoElement::index()",
            "style::gecko::pseudo_element::PseudoElement::index()",
        ),
        (
            "static void alloc::boxed::{{impl}}::call_box<(),closure>(struct closure *, <NoType>)",
            "alloc::boxed::{{impl}}::call_box<(),closure>(struct closure *, <NoType>)",
        ),
        ("static void core::option::expect_failed()", "core::option::expect_failed()"),
        (
            "struct style::gecko_bindings::sugar::ownership::Strong<style::gecko_bindings::structs::root::RawServoStyleSheetContents> geckoservo::glue::Servo_StyleSheet_Empty(style::gecko_bindings::structs::root::mozilla::css::SheetParsingMode)",  # noqa
            "geckoservo::glue::Servo_StyleSheet_Empty(style::gecko_bindings::structs::root::mozilla::css::SheetParsingMode)",  # noqa
        ),
        (
            "unsigned int encoding_glue::mozilla_encoding_encode_from_utf16(struct encoding_rs::Encoding * *, unsigned short *, unsigned int, struct nsstring::nsACString *)",  # noqa
            "encoding_glue::mozilla_encoding_encode_from_utf16(struct encoding_rs::Encoding * *, unsigned short *, unsigned int, struct nsstring::nsACString *)",  # noqa
        ),
        (
            "void geckoservo::glue::Servo_MaybeGCRuleTree(struct style::gecko_bindings::bindings::RawServoStyleSet *)",  # noqa
            "geckoservo::glue::Servo_MaybeGCRuleTree(struct style::gecko_bindings::bindings::RawServoStyleSet *)",  # noqa
        ),
        # Handle whitespace between function and parenthesized arguments and [clone .cold.xxx] correctly
        (
            "[thunk]:CShellItem::QueryInterface`adjustor{12}' (_GUID const&, void**)",
            "[thunk]:CShellItem::QueryInterface`adjustor{12}' (_GUID const&, void**)",
        ),
        (
            "nsXPConnect::InitStatics() [clone .cold.638]",
            "nsXPConnect::InitStatics() [clone .cold.638]",
        ),
        (
            "js::AssertObjectIsSavedFrameOrWrapper(JSContext*, JS::Handle<JSObject*>) [clone .isra.234] [clone .cold.687]",  # noqa
            "js::AssertObjectIsSavedFrameOrWrapper(JSContext*, JS::Handle<JSObject*>) [clone .isra.234] [clone .cold.687]",  # noqa
        ),
        # Handle an aberrant case
        (
            "(anonymous namespace)::EnqueueTask(already_AddRefed<nsIRunnable>, int)",
            "(anonymous namespace)::EnqueueTask(already_AddRefed<nsIRunnable>, int)",
        ),
    ],
)
def test_drop_prefix_and_return_type(function, expected):
    assert drop_prefix_and_return_type(function) == expected


@pytest.mark.parametrize(
    "item, expected",
    [
        ("", None),
        ("foo", None),
        (
            "0b794045-87ec-4649-9ce1-73ec10191120",
            "0b794045-87ec-4649-9ce1-73ec10191120",
        ),
        (
            "bp-0b794045-87ec-4649-9ce1-73ec10191120",
            "0b794045-87ec-4649-9ce1-73ec10191120",
        ),
        (
            "https://crash-stats.mozilla.org/report/index/0b794045-87ec-4649-9ce1-73ec10191120",
            "0b794045-87ec-4649-9ce1-73ec10191120",
        ),
    ],
)
def test_parse_crashid(item, expected):
    assert parse_crashid(item) == expected


@pytest.mark.parametrize(
    "text, expected",
    [
        # Bad data begets itself
        (None, None),
        (16, 16),
        # Good hex things get fixed correctly
        ("0x32ec0", "0x32ec0"),
        ("0x0000000000032ec0", "0x32ec0"),
        # This is a weird case where it looks like a decimal and should probably be
        # a decimal, but our code strips it and puts 0x in front. I'm going to leave
        # it like this for now.
        ("40", "0x40"),
    ],
)
def test_strip_leading_zeros(text, expected):
    assert strip_leading_zeros(text) == expected



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from pathlib import Path
import re

import importlib_resources


# This is a hack because sentinels can be a tuple, with the second item being
# a function to verify if the sentinel applies. It's quite hard to express
# that in a .txt file, so this special value is here. This list should not
# grow much, and if it does, we should find a better solution for handling
# these special values.
_SPECIAL_EXTENDED_VALUES = {
    "signature_sentinels": [
        (
            "mozilla::ipc::RPCChannel::Call(IPC::Message*, IPC::Message*)",
            lambda x: "CrashReporter::CreatePairedMinidumps(void*, unsigned long, nsAString_internal*, nsILocalFile**, nsILocalFile**)"
            in x,  # noqa
        )
    ]
}


class BadRegularExpressionLineError(Exception):
    """Raised when a file contains an invalid regular expression."""


class _IncludedSource:
    def __repr__(self):
        return "INCLUDED"

    def __str__(self):
        return "INCLUDED"


INCLUDED = _IncludedSource()


def get_filepath(name, source=INCLUDED):
    """Build a file path from the

    :param name: the signature list name
    :param source: where to look for the specified signature list file: ``INCLUDED`` to
        look at included signature list files or the directory on the file system
        as a Path or string

    :returns: Path to file in package

    """
    if source is INCLUDED:
        package_name = ".".join(__name__.split(".")[0:-1])
        return importlib_resources.files(package_name).joinpath(f"siglists/{name}.txt")

    source = Path(source)
    return source / f"{name}.txt"


def get_signature_list_content(name, source=INCLUDED):
    """Return a tuple, each value being a line of the source file.

    Remove empty lines and comments (lines starting with a '#').

    :param name: the signature list name
    :param source: where to look for the specified signature list file: ``INCLUDED`` to
        look at included signature list files or the directory on the file system
        as a Path or string

    :returns: tuple of lines from file

    """
    filepath = get_filepath(name, source=source)
    lines = []

    with filepath.open("rb") as fp:
        for i, line in enumerate(fp):
            line = line.decode("utf-8", "strict").strip()
            if not line or line.startswith("#"):
                continue

            try:
                re.compile(line)
            except Exception as exc:
                raise BadRegularExpressionLineError(
                    f"Regex error: {exc} in file {filepath} at line {i}"
                ) from exc

            lines.append(line)

    if source in _SPECIAL_EXTENDED_VALUES:
        lines = lines + _SPECIAL_EXTENDED_VALUES[source]

    return tuple(lines)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import print_function

import argparse
import json
import sys

from .generator import SignatureGenerator
from .utils import convert_to_crash_data


DESCRIPTION = """
Given a signature data structure as JSON via stdin, generates the signature.

If you pass a processed crash, this will convert it.
"""


def main():
    """Takes crash data via stdin and generates a Socorro signature"""
    parser = argparse.ArgumentParser(description=DESCRIPTION)
    parser.add_argument(
        "-v", "--verbose", help="increase output verbosity", action="store_true"
    )
    parser.add_argument(
        "--signature-list-dir",
        required=False,
        help=(
            "directory of signature list files to use; if not specified, uses the "
            + "included signature list files"
        ),
    )
    args = parser.parse_args()

    generator_kwargs = {}
    if args.signature_list_dir:
        generator_kwargs = {
            "signature_list_dir": args.signature_list_dir,
        }
    generator = SignatureGenerator(**generator_kwargs)

    crash_data = json.loads(sys.stdin.read())

    if "json_dump" in crash_data:
        # This is an indicator that the crash data is a processed crash and
        # needs to be converted
        if args.verbose:
            print("Crash data is a processed crash. Converting ...")
        crash_data = convert_to_crash_data(crash_data)

    result = generator.generate(crash_data)
    if args.verbose:
        for item in result.debug_log:
            print(item)

    return_dict = {
        "signature": result.signature,
        "proto_signature": result.extra.get("proto_signature", ""),
        "notes": result.notes,
    }
    print(json.dumps(return_dict, indent=2))



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import importlib
from pathlib import Path

import pytest


# NOTE(willkg): We do this so that we can extract signature generation into its
# own namespace as an external library. This allows the tests to run if it's in
# "siggen" or "socorro.signature".
base_module = ".".join(__name__.split(".")[:-2])
siglists_utils = importlib.import_module(base_module + ".siglists_utils")


class TestSigLists:
    def test_loading_files(self):
        signature_lists = (
            "irrelevant_signature_re",
            "prefix_signature_re",
            "signature_sentinels",
            "signatures_with_line_numbers_re",
        )

        for name in signature_lists:
            content = siglists_utils.get_signature_list_content(name)
            assert len(content) > 0

            for line in content:
                assert len(line) > 0
                # Some items can be tuples; for the str lines, make sure they don't
                # start with a #
                if isinstance(line, str):
                    assert not line.startswith("#")

    def test_valid_entries(self):
        source = Path(__file__).parent / "siglists"
        expected = ("fooBarStuff", "moz::.*", "@0x[0-9a-fA-F]{2,}")
        content = siglists_utils.get_signature_list_content(
            "test-valid-sig-list", source=source
        )
        assert content == expected

    def test_invalid_entry(self):
        source = Path(__file__).parent / "siglists"

        with pytest.raises(siglists_utils.BadRegularExpressionLineError) as exc_info:
            siglists_utils.get_signature_list_content(
                "test-invalid-sig-list", source=source
            )

        msg = exc_info.exconly()
        assert "BadRegularExpressionLineError: Regex error: " in msg
        assert msg.endswith("at line 3")
        assert "test-invalid-sig-list.txt" in msg



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import copy
import importlib
import json
import re

import pytest


# NOTE(willkg): We do this so that we can extract signature generation into its
# own namespace as an external library. This allows the tests to run if it's in
# "siggen" or "socorro.signature".
BASE_MODULE = ".".join(__name__.split(".")[:-2])
rules = importlib.import_module(BASE_MODULE + ".rules")
generator = importlib.import_module(BASE_MODULE + ".generator")


@pytest.mark.parametrize(
    "frame, expected_function",
    [
        # Don't die on missing data
        ({}, None),
        # If file is missing, it should just return the same frame
        ({"function": "panic"}, "panic"),
        # Fix function if (file, function) are in lookup table
        (
            {
                "file": "git:github.com/rust-lang/rust:src/libstd/panicking.rs:91856ed52c58aa5ba66a015354d1cc69e9779bdf",  # noqa
                "function": "begin_panic_fmt",
            },
            "std::panicking::begin_panic_fmt",
        ),
    ],
)
def test_fix_missing_module(frame, expected_function):
    actual_function = rules.fix_missing_module(frame).get("function")
    assert actual_function == expected_function


class TestCSignatureTool:
    @staticmethod
    def setup_config_c_sig_tool(
        ig=("ignored1",),
        pr=("pre1", "pre2"),
        si=("fnNeedNumber",),
        td=(r"foo32\.dll.*",),
        ss=("sentinel", ("sentinel2", lambda x: "ff" in x)),
    ):
        tool = rules.CSignatureTool()
        tool.irrelevant_signature_re = tool.build_re(ig)
        tool.prefix_signature_re = tool.build_re(pr)
        tool.signatures_with_line_numbers_re = tool.build_re(si)
        tool.signature_sentinels = ss
        return tool

    def test_c_config_tool_init(self):
        """test_C_config_tool_init: constructor test"""
        exp_irrelevant_signature_re = re.compile(r"ignored1")
        exp_prefix_signature_re = re.compile(r"pre1|pre2")
        exp_signatures_with_line_numbers_re = re.compile(r"fnNeedNumber")
        fixup_space = re.compile(r" (?=[\*&,])")
        fixup_comma = re.compile(r",(?! )")

        s = self.setup_config_c_sig_tool()

        assert exp_irrelevant_signature_re.pattern == s.irrelevant_signature_re.pattern
        assert exp_prefix_signature_re.pattern == s.prefix_signature_re.pattern
        assert (
            exp_signatures_with_line_numbers_re.pattern
            == s.signatures_with_line_numbers_re.pattern
        )

        assert fixup_space.pattern == s.fixup_space.pattern
        assert fixup_comma.pattern == s.fixup_comma.pattern

    def test_create_frame_list(self):
        thread_0 = {
            "frames": [
                {
                    "frame": 0,
                    "function": "NtWaitForMultipleObjects",
                    "function_offset": "0x15",
                    "module": "ntdll.dll",
                    "module_offset": "0x2015d",
                    "offset": "0x77ad015d",
                    "trust": "context",
                },
                {
                    "frame": 1,
                    "function": "WaitForMultipleObjectsEx",
                    "function_offset": "0xff",
                    "module": "KERNELBASE.dll",
                    "module_offset": "0x115f6",
                    "offset": "0x775e15f6",
                    "trust": "cfi",
                },
                {
                    "frame": 2,
                    "function": "WaitForMultipleObjectsExImplementation",
                    "function_offset": "0x8d",
                    "module": "kernel32.dll",
                    "module_offset": "0x119f7",
                    "offset": "0x766119f7",
                    "trust": "cfi",
                },
                {
                    "frame": 3,
                    "function": "RealMsgWaitForMultipleObjectsEx",
                    "function_offset": "0xe1",
                    "module": "user32.dll",
                    "module_offset": "0x20869",
                    "offset": "0x77370869",
                    "trust": "cfi",
                },
                {
                    "frame": 4,
                    "function": "MsgWaitForMultipleObjects",
                    "function_offset": "0x1e",
                    "module": "user32.dll",
                    "module_offset": "0x20b68",
                    "offset": "0x77370b68",
                    "trust": "cfi",
                },
                {
                    "file": "F_766591945_______________________________________",
                    "frame": 5,
                    "function": "F_1428703866________________________________",
                    "function_offset": "0xc1",
                    "line": 203,
                    "module": "NPSWF32_14_0_0_125.dll",
                    "module_offset": "0x35bf21",
                    "offset": "0x5e39bf21",
                    "trust": "cfi",
                },
            ],
            "threads_index": 0,
            "frame_count": 32,
        }

        s = self.setup_config_c_sig_tool()
        frame_signatures_list = s.create_frame_list(thread_0)
        expected = [
            "NtWaitForMultipleObjects",
            "WaitForMultipleObjectsEx",
            "WaitForMultipleObjectsExImplementation",
            "RealMsgWaitForMultipleObjectsEx",
            "MsgWaitForMultipleObjects",
            "F_1428703866________________________________",
        ]
        assert frame_signatures_list == expected

    def test_create_frame_list_with_inlines(self):
        thread_0 = {
            "frames": [
                {
                    "frame": 0,
                    "function": "NtWaitForMultipleObjects",
                    "function_offset": "0x15",
                    "module": "ntdll.dll",
                    "module_offset": "0x2015d",
                    "offset": "0x77ad015d",
                    "trust": "context",
                    "inlines": None,
                },
                {
                    "frame": 1,
                    "function": "WaitForMultipleObjectsEx",
                    "function_offset": "0xff",
                    "module": "KERNELBASE.dll",
                    "module_offset": "0x115f6",
                    "offset": "0x775e15f6",
                    "trust": "cfi",
                    "inlines": [
                        {
                            "function": "inline_a",
                            "file": "some_file.rs",
                            "line": 59,
                        },
                        {
                            "function": "inline_b",
                            "file": "some_file.rs",
                            "line": 299,
                        },
                    ],
                },
                {
                    "frame": 2,
                    "function": "WaitForMultipleObjectsExImplementation",
                    "function_offset": "0x8d",
                    "module": "kernel32.dll",
                    "module_offset": "0x119f7",
                    "offset": "0x766119f7",
                    "trust": "cfi",
                },
            ],
            "threads_index": 0,
            "frame_count": 3,
        }

        s = self.setup_config_c_sig_tool()
        frame_signatures_list = s.create_frame_list(thread_0)
        expected = [
            "NtWaitForMultipleObjects",
            "inline_a",
            "inline_b",
            "WaitForMultipleObjectsEx",
            "WaitForMultipleObjectsExImplementation",
        ]
        assert frame_signatures_list == expected

    def test_no_mutation(self):
        s = self.setup_config_c_sig_tool()
        frames = {
            "frames": [
                {
                    "frame": 0,
                    "function": "NtWaitForMultipleObjects",
                    "function_offset": "0x15",
                    "module": "ntdll.dll",
                    "module_offset": "0x2015d",
                    "offset": "0x77ad015d",
                    "trust": "context",
                },
                {
                    "frame": 1,
                    "function": "WaitForMultipleObjectsEx",
                    "function_offset": "0xff",
                    "module": "KERNELBASE.dll",
                    "module_offset": "0x115f6",
                    "offset": "0x775e15f6",
                    "trust": "cfi",
                },
            ]
        }
        frames_copy = copy.deepcopy(frames)
        s = self.setup_config_c_sig_tool()
        s.create_frame_list(frames_copy)

        # If they're the same, then there was no mutation
        assert frames == frames_copy

    @pytest.mark.parametrize(
        "args, expected",
        [
            # module, function, file, line, module_offset, offset, unloaded_modules
            (("module", "", "source/", "23", "0xfff"), "source#23"),
            (("module", "", "source\\", "23", "0xfff"), "source#23"),
            (("module", "", "/a/b/c/source", "23", "0xfff"), "source#23"),
            (("module", "", "\\a\\b\\c\\source", "23", "0xfff"), "source#23"),
            (("module", "", "\\a\\b\\c\\source", "23", "0xfff"), "source#23"),
            (("module", "", "\\a\\b\\c\\source", "", "0xfff"), "module@0xfff"),
            (("module", "", "", "23", "0xfff"), "module@0xfff"),
            (("module", "", "", "", "0xfff"), "module@0xfff"),
            ((None, "", "", "", "0xfff"), "@0xfff"),
            # Make sure frame normalization uses the right function: normalize
            # Rust frame (has a Rust fingerprint)
            (
                (
                    "module",
                    "expect_failed::h7f635057bfba806a",
                    "hg:hg.mozilla.org/a/b:servio/wrapper.rs:44444444444",
                    "23",
                    "0xfff",
                ),
                "expect_failed",
            ),
            # Handle normalization with unloaded modules
            (
                (
                    "",
                    "",
                    "",
                    "",
                    "",
                    "",
                    [{"module": "unmod"}],
                ),
                "(unloaded unmod)",
            ),
            (
                (
                    "",
                    "",
                    "",
                    "",
                    "",
                    "",
                    [{"module": "unmod", "offsets": ["0x0000000000005387"]}],
                ),
                "(unloaded unmod@0x5387)",
            ),
        ],
    )
    def test_normalize_frame(self, args, expected):
        """test_normalize: bunch of variations"""
        s = self.setup_config_c_sig_tool()
        ret = s.normalize_frame(*args)
        assert expected == ret

    @pytest.mark.parametrize(
        "function, line, expected",
        [
            # Verify function and line number handling
            ("fn", "23", "fn"),
            ("fnNeedNumber", "23", "fnNeedNumber:23"),
            # Remove function arguments
            ("f( *s)", "23", "f"),
            ("f( &s)", "23", "f"),
            ("f( *s , &n)", "23", "f"),
            ("f3(s,t,u)", "23", "f3"),
            ("operator()(s,t,u)", "23", "operator()"),
            ("::(anonymous namespace)::f3(s,t,u)", "23", "::(anonymous namespace)::f3"),
            (
                "mozilla::layers::D3D11YCbCrImage::GetAsSourceSurface()",
                "23",
                "mozilla::layers::D3D11YCbCrImage::GetAsSourceSurface",
            ),
            (
                "mozilla::layers::BasicImageLayer::Paint(mozilla::gfx::DrawTarget*, mozilla::gfx::PointTyped<mozilla::gfx::UnknownUnits, float> const&, mozilla::layers::Layer*)",
                "23",  # noqa
                "mozilla::layers::BasicImageLayer::Paint",
            ),
            (
                "void nsDocumentViewer::DestroyPresShell()",
                "23",
                "nsDocumentViewer::DestroyPresShell",
            ),
            (
                "bool CCGraphBuilder::BuildGraph(class js::SliceBudget& const)",
                "23",
                "CCGraphBuilder::BuildGraph",
            ),
            # Handle converting types to generic
            ("f<3>(s,t,u)", "23", "f<T>"),
            ("Alpha<Bravo<Charlie>, Delta>::Echo<Foxtrot>", "23", "Alpha<T>::Echo<T>"),
            (
                "thread_start<unsigned int (__cdecl*)(void* __ptr64)>",
                "23",
                "thread_start<T>",
            ),
            # Handle prefixes and return types
            (
                "class JSObject* DoCallback<JSObject*>(class JS::CallbackTracer*, class JSObject**, const char*)",
                "23",  # noqa
                "DoCallback<T>",
            ),
            # But don't run drop_prefix_and_return_types for operator overloading
            # functions
            (
                "JS::Heap<JSObject*>::operator JSObject* const &()",
                "23",
                "JS::Heap<T>::operator JSObject* const&",
            ),
            # Drop cv/ref qualifiers at end
            (
                "JSObject::allocKindForTenure const",
                "23",
                "JSObject::allocKindForTenure",
            ),
            (
                "mozilla::jni::GlobalRef<mozilla::jni::Object>::operator=(mozilla::jni::Ref<mozilla::jni::Object, _jobject*> const&)&",
                "23",  # noqa
                "mozilla::jni::GlobalRef<T>::operator=",
            ),
            (
                "mozilla::Maybe<unsigned long>::value() const &",
                "23",  # noqa
                "mozilla::Maybe<T>::value",
            ),
            # Normalize anonymous namespace
            ("`anonymous namespace'::foo", "23", "(anonymous namespace)::foo"),
            ("(anonymous namespace)::foo", "23", "(anonymous namespace)::foo"),
            # Normalize lambda numbers
            (
                "ShutdownWorkThreads::$_52::__invoke",
                "23",
                "ShutdownWorkThreads::$::__invoke",
            ),
        ],
    )
    def test_normalize_cpp_function(self, function, line, expected):
        """Test normalization for cpp functions"""
        s = self.setup_config_c_sig_tool()
        assert s.normalize_cpp_function(function, line) == expected

    @pytest.mark.parametrize(
        "function, line, expected",
        [
            # Verify removal of fingerprints
            ("expect_failed::h7f635057bfba806a", "23", "expect_failed"),
            ("expect_failed::h7f6350::blah", "23", "expect_failed::h7f6350::blah"),
            # Handle prefixes, return types, types, and traits
            (
                "static void servo_arc::Arc<style::gecko_properties::ComputedValues>::drop_slow<style::gecko_properties::ComputedValues>()",
                "23",  # noqa
                "servo_arc::Arc<T>::drop_slow<T>",
            ),
            (
                "static void core::ptr::drop_in_place<style::stylist::CascadeData>(struct style::stylist::CascadeData*)",
                "23",  # noqa
                "core::ptr::drop_in_place<T>",
            ),
            # Handle trait methods by not collapsing them
            (
                "<rayon_core::job::HeapJob<BODY> as rayon_core::job::Job>::execute",
                "23",
                "<rayon_core::job::HeapJob<BODY> as rayon_core::job::Job>::execute",
            ),
        ],
    )
    def test_normalize_rust_function(self, function, line, expected):
        """Test normalization for Rust functions"""
        s = self.setup_config_c_sig_tool()
        assert s.normalize_rust_function(function, line) == expected

    def test_generate_1(self):
        """test_generate_1: simple"""
        s = self.setup_config_c_sig_tool(ig=["a", "b", "c"], pr=["d", "e", "f"])
        a = list("abcdefghijklmnopqrstuvwxyz")
        sig, notes, debug_notes = s.generate(a)
        assert sig == "d | e | f | g"

        a = list("abcdaeafagahijklmnopqrstuvwxyz")
        sig, notes, debug_notes = s.generate(a)
        assert sig == "d | e | f | g"

    def test_generate_2a(self):
        """test_generate_2a: way too long"""
        s = self.setup_config_c_sig_tool(ig=["a", "b", "c"], pr=["d", "e", "f"])
        a = list("abcdefghijklmnopqrstuvwxyz")
        a[3] = a[3] * 70
        a[4] = a[4] * 70
        a[5] = a[5] * 70
        a[6] = a[6] * 70
        a[7] = a[7] * 70
        expected = (
            "dddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddd | "
            "eeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeee | "
            "ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff | "
            "gggggggggggggggggggggggggggggggggggggggggggggggggggggggggggggggggggggg"
        )
        sig, notes, debug_notes = s.generate(a)
        assert sig == expected

    def test_generate_3(self):
        """test_generate_3: simple sentinel"""
        s = self.setup_config_c_sig_tool(ig=["a", "b", "c"], pr=["d", "e", "f"])
        a = list("abcdefghabcfaeabdijklmnopqrstuvwxyz")
        a[7] = "sentinel"
        sig, notes, debug_notes = s.generate(a)
        assert sig == "sentinel"

        s = self.setup_config_c_sig_tool(
            ig=["a", "b", "c", "sentinel"], pr=["d", "e", "f"]
        )
        sig, notes, debug_notes = s.generate(a)
        assert sig == "f | e | d | i"

    def test_generate_4(self):
        """test_generate_4: tuple sentinel"""
        s = self.setup_config_c_sig_tool(ig=["a", "b", "c"], pr=["d", "e", "f"])
        a = list("abcdefghabcfaeabdijklmnopqrstuvwxyz")
        a[7] = "sentinel2"
        sig, notes, debug_notes = s.generate(a)
        assert sig == "d | e | f | g"

        s = self.setup_config_c_sig_tool(ig=["a", "b", "c"], pr=["d", "e", "f"])
        a = list("abcdefghabcfaeabdijklmnopqrstuvwxyz")
        a[7] = "sentinel2"
        a[22] = "ff"
        sig, notes, debug_notes = s.generate(a)
        assert sig == "sentinel2"

        s = self.setup_config_c_sig_tool(
            ig=["a", "b", "c", "sentinel2"], pr=["d", "e", "f"]
        )
        a = list("abcdefghabcfaeabdijklmnopqrstuvwxyz")
        a[7] = "sentinel2"
        a[22] = "ff"
        sig, notes, debug_notes = s.generate(a)
        assert sig == "f | e | d | i"

    def test_generate_with_merged_dll(self):
        sig_tool = self.setup_config_c_sig_tool(ig=["a", "b", "c"], pr=["d", "e", "f"])
        source_list = [
            "a",
            "d",
            "foo32.dll@0x231423",
            "foo32.dll",
            "foo32.dll@0x42",
            "g",
        ]
        sig, notes, debug_notes = sig_tool.generate(source_list)
        assert sig == "d | foo32.dll | g"

        source_list = ["foo32.dll", "foo32.dll@0x231423", "g"]
        sig, notes, debug_notes = sig_tool.generate(source_list)
        assert sig == "foo32.dll | g"


class TestJavaSignatureTool:
    def test_bad_stack(self):
        j = rules.JavaSignatureTool()
        java_stack_trace = 17
        sig, notes, debug_notes = j.generate(java_stack_trace, delimiter=": ")
        assert sig == "EMPTY: Java stack trace not in expected format"
        assert notes == ["JavaSignatureTool: stack trace not in expected format"]

    def test_basic_stack_frame_with_line_number(self):
        j = rules.JavaSignatureTool()
        java_stack_trace = (
            "SomeJavaException: totally made up  \n"
            "at org.mozilla.lars.myInvention(larsFile.java:666)"
        )
        sig, notes, debug_notes = j.generate(java_stack_trace, delimiter=": ")
        e = "SomeJavaException: totally made up at org.mozilla.lars.myInvention(larsFile.java)"
        assert sig == e
        assert notes == []

    def test_basic_stack_frame(self):
        j = rules.JavaSignatureTool()
        java_stack_trace = (
            "SomeJavaException: totally made up  \n"
            "at org.mozilla.lars.myInvention(larsFile.java)"
        )
        sig, notes, debug_notes = j.generate(java_stack_trace, delimiter=": ")
        e = "SomeJavaException: totally made up at org.mozilla.lars.myInvention(larsFile.java)"
        assert sig == e
        assert notes == []

    def test_long_exception_description(self):
        j = rules.JavaSignatureTool()
        java_stack_trace = (
            "   SomeJavaException: %s \nat org.mozilla.lars.myInvention(larsFile.java)"
            % ("t" * 1000)
        )
        sig, notes, debug_notes = j.generate(java_stack_trace, delimiter=": ")
        expected = "SomeJavaException: at org.mozilla.lars.myInvention(larsFile.java)"
        assert sig == expected
        expected = [
            "JavaSignatureTool: dropped Java exception description due to length"
        ]
        assert notes == expected

    def test_long_exception_description_with_line_number(self):
        j = rules.JavaSignatureTool()
        java_stack_trace = (
            "   SomeJavaException: %s  \nat org.mozilla.lars.myInvention(larsFile.java:1234)"
            % ("t" * 1000)
        )
        sig, notes, debug_notes = j.generate(java_stack_trace, delimiter=": ")
        expected = "SomeJavaException: at org.mozilla.lars.myInvention(larsFile.java)"
        assert sig == expected
        expected = [
            "JavaSignatureTool: dropped Java exception description due to length"
        ]
        assert notes == expected

    def test_no_description(self):
        j = rules.JavaSignatureTool()
        java_stack_trace = (
            "   SomeJavaException\n"
            "at org.mozilla.lars.myInvention(larsFile.java:1234)"
        )
        sig, notes, debug_notes = j.generate(java_stack_trace, delimiter=": ")
        e = "SomeJavaException: at org.mozilla.lars.myInvention(larsFile.java)"
        assert sig == e
        e = ["JavaSignatureTool: stack trace line 1 is not in the expected format"]
        assert notes == e

    def test_frame_with_line_ending_but_missing_second_line(self):
        j = rules.JavaSignatureTool()
        java_stack_trace = "SomeJavaException: totally made up  \n"
        sig, notes, debug_notes = j.generate(java_stack_trace, delimiter=": ")
        e = "SomeJavaException: totally made up"
        assert sig == e
        e = ["JavaSignatureTool: stack trace line 2 is missing"]
        assert notes == e

    def test_frame_missing_second_line(self):
        j = rules.JavaSignatureTool()
        java_stack_trace = "SomeJavaException: totally made up  "
        sig, notes, debug_notes = j.generate(java_stack_trace, delimiter=": ")
        e = "SomeJavaException: totally made up"
        assert sig == e
        e = ["JavaSignatureTool: stack trace line 2 is missing"]
        assert notes == e

    def test_frame_with_leading_whitespace(self):
        j = rules.JavaSignatureTool()
        java_stack_trace = (
            "   SomeJavaException: totally made up  \n"
            "at org.mozilla.lars.myInvention("
            "foolarsFile.java:1234)"
        )
        sig, notes, debug_notes = j.generate(java_stack_trace, delimiter=": ")
        expected = "SomeJavaException: totally made up at org.mozilla.lars.myInvention(foolarsFile.java)"
        assert sig == expected

    def test_no_interference(self):
        # In general addresses of the form ``@xxxxxxxx`` are to be replaced
        # with the literal ``<addr>``, however in this case, the hex address is
        # not in the expected location and should therefore be left alone
        j = rules.JavaSignatureTool()
        java_stack_trace = (
            "SomeJavaException: totally made up  \n"
            "at org.mozilla.lars.myInvention(larsFile.java:@abef1234)"
        )
        sig, notes, debug_notes = j.generate(java_stack_trace, delimiter=" ")
        e = (
            "SomeJavaException totally made up "
            "at org.mozilla.lars.myInvention("
            "larsFile.java:@abef1234)"
        )
        assert sig == e
        assert notes == []

    def test_replace_address(self):
        j = rules.JavaSignatureTool()
        java_stack_trace = """
java.lang.IllegalArgumentException: Given view not a child of android.widget.AbsoluteLayout@4054b560
\tat android.view.ViewGroup.updateViewLayout(ViewGroup.java:1968)
\tat org.mozilla.gecko.GeckoApp.repositionPluginViews(GeckoApp.java:1492)
\tat org.mozilla.gecko.GeckoApp.repositionPluginViews(GeckoApp.java:1475)
\tat org.mozilla.gecko.gfx.LayerController$2.run(LayerController.java:269)
\tat android.os.Handler.handleCallback(Handler.java:587)
\tat android.os.Handler.dispatchMessage(Handler.java:92)
\tat android.os.Looper.loop(Looper.java:150)
\tat org.mozilla.gecko.GeckoApp$32.run(GeckoApp.java:1670)
\tat android.os.Handler.handleCallback(Handler.java:587)
\tat android.os.Handler.dispatchMessage(Handler.java:92)
\tat android.os.Looper.loop(Looper.java:150)
\tat android.app.ActivityThread.main(ActivityThread.java:4293)
\tat java.lang.reflect.Method.invokeNative(Native Method)
\tat java.lang.reflect.Method.invoke(Method.java:507)
\tat com.android.internal.os.ZygoteInit$MethodAndArgsCaller.run(ZygoteInit.java:849)
\tat com.android.internal.os.ZygoteInit.main(ZygoteInit.java:607)
\tat dalvik.system.NativeStart.main(Native Method)""".lstrip()
        sig, notes, debug_notes = j.generate(java_stack_trace, delimiter=": ")
        e = (
            "java.lang.IllegalArgumentException: "
            "Given view not a child of android.widget.AbsoluteLayout@<addr>: "
            "at android.view.ViewGroup.updateViewLayout(ViewGroup.java)"
        )
        assert sig == e
        assert notes == []

    def test_replace_address_with_trailing_text(self):
        j = rules.JavaSignatureTool()
        java_stack_trace = """
android.view.WindowManager$BadTokenException: Unable to add window -- token android.os.BinderProxy@406237c0 is not valid; is your activity running?
\tat android.view.ViewRoot.setView(ViewRoot.java:533)
\tat android.view.WindowManagerImpl.addView(WindowManagerImpl.java:202)
\tat android.view.WindowManagerImpl.addView(WindowManagerImpl.java:116)
\tat android.view.Window$LocalWindowManager.addView(Window.java:424)
\tat android.app.ActivityThread.handleResumeActivity(ActivityThread.java:2174)
\tat android.app.ActivityThread.handleLaunchActivity(ActivityThread.java:1672)
\tat android.app.ActivityThread.access$1500(ActivityThread.java:117)
\tat android.app.ActivityThread$H.handleMessage(ActivityThread.java:935)
\tat android.os.Handler.dispatchMessage(Handler.java:99)
\tat android.os.Looper.loop(Looper.java:130)
\tat android.app.ActivityThread.main(ActivityThread.java:3687)
\tat java.lang.reflect.Method.invokeNative(Native Method)
\tat java.lang.reflect.Method.invoke(Method.java:507)
\tat com.android.internal.os.ZygoteInit$MethodAndArgsCaller.run(ZygoteInit.java:867)
\tat com.android.internal.os.ZygoteInit.main(ZygoteInit.java:625)
\tat dalvik.system.NativeStart.main(Native Method)
""".lstrip()  # noqa
        sig, notes, debug_notes = j.generate(java_stack_trace, delimiter=": ")
        e = (
            "android.view.WindowManager$BadTokenException: "
            "Unable to add window -- token android.os.BinderProxy@<addr> "
            "is not valid; is your activity running? "
            "at android.view.ViewRoot.setView(ViewRoot.java)"
        )
        assert sig == e
        assert notes == []

    def test_replace_address_trailing_whitespace(self):
        j = rules.JavaSignatureTool()
        java_stack_trace = """
java.lang.IllegalArgumentException: Receiver not registered: org.mozilla.gecko.GeckoNetworkManager@405afea8
\tat android.app.LoadedApk.forgetReceiverDispatcher(LoadedApk.java:610)
\tat android.app.ContextImpl.unregisterReceiver(ContextImpl.java:883)
\tat android.content.ContextWrapper.unregisterReceiver(ContextWrapper.java:331)
\tat org.mozilla.gecko.GeckoNetworkManager.stopListening(GeckoNetworkManager.java:141)
\tat org.mozilla.gecko.GeckoNetworkManager.stop(GeckoNetworkManager.java:136)
\tat org.mozilla.gecko.GeckoApp.onApplicationPause(GeckoApp.java:2130)
\tat org.mozilla.gecko.GeckoApplication.onActivityPause(GeckoApplication.java:55)
\tat org.mozilla.gecko.GeckoActivity.onPause(GeckoActivity.java:22)
\tat org.mozilla.gecko.GeckoApp.onPause(GeckoApp.java:1948)
\tat android.app.Activity.performPause(Activity.java:3877)
\tat android.app.Instrumentation.callActivityOnPause(Instrumentation.java:1191)
\tat android.app.ActivityThread.performPauseActivity(ActivityThread.java:2345)
\tat android.app.ActivityThread.performPauseActivity(ActivityThread.java:2315)
\tat android.app.ActivityThread.handlePauseActivity(ActivityThread.java:2295)
\tat android.app.ActivityThread.access$1700(ActivityThread.java:117)
\tat android.app.ActivityThread$H.handleMessage(ActivityThread.java:942)
\tat android.os.Handler.dispatchMessage(Handler.java:99)
\tat android.os.Looper.loop(Looper.java:130)
\tat android.app.ActivityThread.main(ActivityThread.java:3691)
\tat java.lang.reflect.Method.invokeNative(Native Method)
\tat java.lang.reflect.Method.invoke(Method.java:507)
\tat com.android.internal.os.ZygoteInit$MethodAndArgsCaller.run(ZygoteInit.java:907)
\tat com.android.internal.os.ZygoteInit.main(ZygoteInit.java:665)
\tat dalvik.system.NativeStart.main(Native Method)
""".lstrip()  # noqa
        sig, notes, debug_notes = j.generate(java_stack_trace, delimiter=": ")
        e = (
            "java.lang.IllegalArgumentException: "
            "Receiver not registered: "
            "org.mozilla.gecko.GeckoNetworkManager@<addr>: "
            "at android.app.LoadedApk.forgetReceiverDispatcher(LoadedApk.java)"
        )
        assert sig == e
        assert notes == []


class TestJavaExceptionSignatureTool:
    @pytest.mark.parametrize(
        "value",
        [
            None,
            "abcde",
            [],
            {},
        ],
    )
    def test_bad_data(self, value):
        tool = rules.JavaExceptionSignatureTool()
        signature, notes, debug_notes = tool.generate(value)
        assert signature == "EMPTY: Java exception not in expected format"
        assert notes == ["JavaExceptionSignatureTool: value not in expected format"]

    @pytest.mark.parametrize(
        "frame, expected",
        [
            ({}, "unknown"),
            # Everything that can be used
            (
                {
                    "module": "org.mozilla.fenix.ext.AddonManagerKt",
                    "function": "getFenixAddons",
                    "filename": "AddonManager.kt",
                },
                "org.mozilla.fenix.ext.AddonManagerKt.getFenixAddons(AddonManager.kt)",
            ),
            # Variations on the module
            (
                {
                    "module": "",
                    "function": "getFenixAddons",
                    "filename": "AddonManager.kt",
                },
                "getFenixAddons(AddonManager.kt)",
            ),
            (
                {
                    "module": None,
                    "function": "getFenixAddons",
                    "filename": "AddonManager.kt",
                },
                "getFenixAddons(AddonManager.kt)",
            ),
            (
                {
                    "function": "getFenixAddons",
                    "filename": "AddonManager.kt",
                },
                "getFenixAddons(AddonManager.kt)",
            ),
            # Variations on function
            (
                {
                    "module": "org.mozilla.fenix.ext.AddonManagerKt",
                    "function": "",
                    "filename": "AddonManager.kt",
                },
                "org.mozilla.fenix.ext.AddonManagerKt(AddonManager.kt)",
            ),
            (
                {
                    "module": "org.mozilla.fenix.ext.AddonManagerKt",
                    "function": None,
                    "filename": "AddonManager.kt",
                },
                "org.mozilla.fenix.ext.AddonManagerKt(AddonManager.kt)",
            ),
            (
                {
                    "module": "org.mozilla.fenix.ext.AddonManagerKt",
                    "filename": "AddonManager.kt",
                },
                "org.mozilla.fenix.ext.AddonManagerKt(AddonManager.kt)",
            ),
            # Variations on filename
            (
                {
                    "module": "org.mozilla.fenix.ext.AddonManagerKt",
                    "function": "getFenixAddons",
                    "filename": "",
                },
                "org.mozilla.fenix.ext.AddonManagerKt.getFenixAddons",
            ),
            (
                {
                    "module": "org.mozilla.fenix.ext.AddonManagerKt",
                    "function": "getFenixAddons",
                    "filename": None,
                },
                "org.mozilla.fenix.ext.AddonManagerKt.getFenixAddons",
            ),
            (
                {
                    "module": "org.mozilla.fenix.ext.AddonManagerKt",
                    "function": "getFenixAddons",
                },
                "org.mozilla.fenix.ext.AddonManagerKt.getFenixAddons",
            ),
        ],
    )
    def test_normalize_frame(self, frame, expected):
        tool = rules.JavaExceptionSignatureTool()
        assert tool.normalize_frame(frame) == expected

    @pytest.mark.parametrize(
        "stacktrace, expected",
        [
            (
                {
                    "stacktrace": {
                        "frames": [
                            {
                                "module": "org.mozilla.fenix.ext.AddonManagerKt",
                                "function": "getFenixAddons",
                                "filename": "AddonManager.kt",
                            },
                        ],
                        "type": "IOException",
                        "module": "java.io",
                    },
                },
                (
                    "java.io.IOException: at "
                    + "org.mozilla.fenix.ext.AddonManagerKt.getFenixAddons(AddonManager.kt)"
                ),
            ),
            (
                {
                    "stacktrace": {
                        "frames": [],
                        "type": "RemoteException",
                        "module": "android.os",
                    },
                },
                "android.os.RemoteException",
            ),
        ],
    )
    def test_generate(self, stacktrace, expected):
        java_exception = {"exception": {"values": [stacktrace]}}
        tool = rules.JavaExceptionSignatureTool()
        signature, notes, debug_notes = tool.generate(java_exception)
        assert signature == expected


#  rules testing section

FRAMES_FROM_JSON_DUMP = {
    "frames": [
        {
            "frame": 0,
            "function": "NtWaitForMultipleObjects",
            "function_offset": "0x15",
            "module": "ntdll.dll",
            "module_offset": "0x2015d",
            "offset": "0x77ad015d",
            "registers": {
                "eax": "0x00000040",
                "ebp": "0x0025e968",
                "ebx": "0x0025e91c",
                "ecx": "0x00000000",
                "edi": "0x00000000",
                "edx": "0x00000000",
                "efl": "0x00200246",
                "eip": "0x77ad015d",
                "esi": "0x00000004",
                "esp": "0x0025e8cc",
            },
            "trust": "context",
        },
        {
            "frame": 1,
            "function": "WaitForMultipleObjectsEx",
            "function_offset": "0xff",
            "module": "KERNELBASE.dll",
            "module_offset": "0x115f6",
            "offset": "0x775e15f6",
            "trust": "cfi",
        },
        {
            "frame": 2,
            "function": "WaitForMultipleObjectsExImplementation",
            "function_offset": "0x8d",
            "module": "kernel32.dll",
            "module_offset": "0x119f7",
            "offset": "0x766119f7",
            "trust": "cfi",
        },
        {
            "frame": 3,
            "function": "RealMsgWaitForMultipleObjectsEx",
            "function_offset": "0xe1",
            "module": "user32.dll",
            "module_offset": "0x20869",
            "offset": "0x77370869",
            "trust": "cfi",
        },
        {
            "frame": 4,
            "function": "MsgWaitForMultipleObjects",
            "function_offset": "0x1e",
            "module": "user32.dll",
            "module_offset": "0x20b68",
            "offset": "0x77370b68",
            "trust": "cfi",
        },
        {
            "file": "F117835525________________________________________",
            "frame": 5,
            "function": "F_1152915508__________________________________",
            "function_offset": "0xbb",
            "line": 118,
            "module": "NPSWF32_14_0_0_125.dll",
            "module_offset": "0x36a13b",
            "offset": "0x5e3aa13b",
            "trust": "cfi",
        },
        {
            "file": "F_851861807_______________________________________",
            "frame": 6,
            "function": "F2166389______________________________________",
            "function_offset": "0xe5",
            "line": 552,
            "module": "NPSWF32_14_0_0_125.dll",
            "module_offset": "0x35faf5",
            "offset": "0x5e39faf5",
            "trust": "cfi",
        },
        {
            "file": "F_851861807_______________________________________",
            "frame": 7,
            "function": "F_917831355___________________________________",
            "function_offset": "0x29b",
            "line": 488,
            "module": "NPSWF32_14_0_0_125.dll",
            "module_offset": "0x360a7b",
            "offset": "0x5e3a0a7b",
            "trust": "cfi",
        },
        {
            "file": "F_851861807_______________________________________",
            "frame": 8,
            "function": "F1315696776________________________________",
            "function_offset": "0xd",
            "line": 439,
            "module": "NPSWF32_14_0_0_125.dll",
            "module_offset": "0x35e2fd",
            "offset": "0x5e39e2fd",
            "trust": "cfi",
        },
        {
            "file": "F_766591945_______________________________________",
            "frame": 9,
            "function": "F_1428703866________________________________",
            "function_offset": "0xc1",
            "line": 203,
            "module": "NPSWF32_14_0_0_125.dll",
            "module_offset": "0x35bf21",
            "offset": "0x5e39bf21",
            "trust": "cfi",
        },
    ],
    "threads_index": 0,
    "frame_count": 32,
}

FRAMES_FROM_JSON_DUMP_WITH_TEMPLATES = {
    "frames": [
        {
            "frame": 0,
            "function": "NtWaitForMultipleObjects",
            "function_offset": "0x15",
            "module": "ntdll.dll",
            "module_offset": "0x2015d",
            "offset": "0x77ad015d",
            "registers": {
                "eax": "0x00000040",
                "ebp": "0x0025e968",
                "ebx": "0x0025e91c",
                "ecx": "0x00000000",
                "edi": "0x00000000",
                "edx": "0x00000000",
                "efl": "0x00200246",
                "eip": "0x77ad015d",
                "esi": "0x00000004",
                "esp": "0x0025e8cc",
            },
            "trust": "context",
        },
        {
            "frame": 1,
            "function": "Alpha<Bravo<Charlie>, Delta>::Echo<Foxtrot>",
            "function_offset": "0xff",
            "module": "KERNELBASE.dll",
            "module_offset": "0x115f6",
            "offset": "0x775e15f6",
            "trust": "cfi",
        },
        {
            "frame": 2,
            "function": "WaitForMultipleObjectsExImplementation",
            "function_offset": "0x8d",
            "module": "kernel32.dll",
            "module_offset": "0x119f7",
            "offset": "0x766119f7",
            "trust": "cfi",
        },
        {
            "frame": 3,
            "function": "RealMsgWaitForMultipleObjectsEx(void **fakeargs)",
            "function_offset": "0xe1",
            "module": "user32.dll",
            "module_offset": "0x20869",
            "offset": "0x77370869",
            "trust": "cfi",
        },
        {
            "frame": 4,
            "function": "MsgWaitForMultipleObjects",
            "function_offset": "0x1e",
            "module": "user32.dll",
            "module_offset": "0x20b68",
            "offset": "0x77370b68",
            "trust": "cfi",
        },
        {
            "file": "F117835525________________________________________",
            "frame": 5,
            "function": "F_1152915508__________________________________",
            "function_offset": "0xbb",
            "line": 118,
            "module": "NPSWF32_14_0_0_125.dll",
            "module_offset": "0x36a13b",
            "offset": "0x5e3aa13b",
            "trust": "cfi",
        },
        {
            "file": "F_851861807_______________________________________",
            "frame": 6,
            "function": "F2166389______________________________________",
            "function_offset": "0xe5",
            "line": 552,
            "module": "NPSWF32_14_0_0_125.dll",
            "module_offset": "0x35faf5",
            "offset": "0x5e39faf5",
            "trust": "cfi",
        },
        {
            "file": "F_851861807_______________________________________",
            "frame": 7,
            "function": "F_917831355___________________________________",
            "function_offset": "0x29b",
            "line": 488,
            "module": "NPSWF32_14_0_0_125.dll",
            "module_offset": "0x360a7b",
            "offset": "0x5e3a0a7b",
            "trust": "cfi",
        },
        {
            "file": "F_851861807_______________________________________",
            "frame": 8,
            "function": "F1315696776________________________________",
            "function_offset": "0xd",
            "line": 439,
            "module": "NPSWF32_14_0_0_125.dll",
            "module_offset": "0x35e2fd",
            "offset": "0x5e39e2fd",
            "trust": "cfi",
        },
        {
            "file": "F_766591945_______________________________________",
            "frame": 9,
            "function": "F_1428703866________________________________",
            "function_offset": "0xc1",
            "line": 203,
            "module": "NPSWF32_14_0_0_125.dll",
            "module_offset": "0x35bf21",
            "offset": "0x5e39bf21",
            "trust": "cfi",
        },
    ],
    "threads_index": 0,
}

FRAMES_FROM_JSON_DUMP_WITH_TEMPLATES_AND_SPECIAL_CASE = {
    "frames": [
        {
            "frame": 0,
            "function": "NtWaitForMultipleObjects",
            "function_offset": "0x15",
            "module": "ntdll.dll",
            "module_offset": "0x2015d",
            "offset": "0x77ad015d",
            "registers": {
                "eax": "0x00000040",
                "ebp": "0x0025e968",
                "ebx": "0x0025e91c",
                "ecx": "0x00000000",
                "edi": "0x00000000",
                "edx": "0x00000000",
                "efl": "0x00200246",
                "eip": "0x77ad015d",
                "esi": "0x00000004",
                "esp": "0x0025e8cc",
            },
            "trust": "context",
        },
        {
            "frame": 1,
            "function": "<name omitted>",
            "function_offset": "0xff",
            "module": "KERNELBASE.dll",
            "module_offset": "0x115f6",
            "offset": "0x775e15f6",
            "trust": "cfi",
        },
        {
            "frame": 2,
            "function": "IPC::ParamTraits<mozilla::net::NetAddr>::Write",
            "function_offset": "0x8d",
            "module": "kernel32.dll",
            "module_offset": "0x119f7",
            "offset": "0x766119f7",
            "trust": "cfi",
        },
        {
            "frame": 3,
            "function": "RealMsgWaitForMultipleObjectsEx(void **fakeargs)",
            "function_offset": "0xe1",
            "module": "user32.dll",
            "module_offset": "0x20869",
            "offset": "0x77370869",
            "trust": "cfi",
        },
        {
            "frame": 4,
            "function": "MsgWaitForMultipleObjects",
            "function_offset": "0x1e",
            "module": "user32.dll",
            "module_offset": "0x20b68",
            "offset": "0x77370b68",
            "trust": "cfi",
        },
        {
            "file": "F117835525________________________________________",
            "frame": 5,
            "function": "F_1152915508__________________________________",
            "function_offset": "0xbb",
            "line": 118,
            "module": "NPSWF32_14_0_0_125.dll",
            "module_offset": "0x36a13b",
            "offset": "0x5e3aa13b",
            "trust": "cfi",
        },
        {
            "file": "F_851861807_______________________________________",
            "frame": 6,
            "function": "F2166389______________________________________",
            "function_offset": "0xe5",
            "line": 552,
            "module": "NPSWF32_14_0_0_125.dll",
            "module_offset": "0x35faf5",
            "offset": "0x5e39faf5",
            "trust": "cfi",
        },
        {
            "file": "F_851861807_______________________________________",
            "frame": 7,
            "function": "F_917831355___________________________________",
            "function_offset": "0x29b",
            "line": 488,
            "module": "NPSWF32_14_0_0_125.dll",
            "module_offset": "0x360a7b",
            "offset": "0x5e3a0a7b",
            "trust": "cfi",
        },
        {
            "file": "F_851861807_______________________________________",
            "frame": 8,
            "function": "F1315696776________________________________",
            "function_offset": "0xd",
            "line": 439,
            "module": "NPSWF32_14_0_0_125.dll",
            "module_offset": "0x35e2fd",
            "offset": "0x5e39e2fd",
            "trust": "cfi",
        },
        {
            "file": "F_766591945_______________________________________",
            "frame": 9,
            "function": "F_1428703866________________________________",
            "function_offset": "0xc1",
            "line": 203,
            "module": "NPSWF32_14_0_0_125.dll",
            "module_offset": "0x35bf21",
            "offset": "0x5e39bf21",
            "trust": "cfi",
        },
    ],
    "threads_index": 0,
}


class TestSignatureGenerationRule:
    def test_java_stack_trace(self):
        sgr = rules.SignatureGenerationRule()

        crash_data = {
            "java_stack_trace": (
                "   SomeJavaException: %s  \nat org.mozilla.lars.myInvention(larsFile.java)"
                % ("t" * 1000)
            )
        }

        result = generator.Result()

        # the call to be tested
        assert sgr.action(crash_data, result) is True

        expected = "SomeJavaException: at org.mozilla.lars.myInvention(larsFile.java)"
        assert result.signature == expected
        assert "proto_signature" not in result.extra
        assert "normalized_frames" not in result.extra
        assert result.notes == [
            "SignatureGenerationRule: JavaSignatureTool: dropped Java exception description due to length"  # noqa
        ]

    def test_c_stack_trace(self):
        sgr = rules.SignatureGenerationRule()

        crash_data = {"os": "Windows NT", "threads": [FRAMES_FROM_JSON_DUMP]}
        result = generator.Result()

        # the call to be tested
        assert sgr.action(crash_data, result) is True

        expected = (
            "MsgWaitForMultipleObjects | F_1152915508__________________________________"
        )
        assert result.signature == expected

        expected_normalized_frames = [
            "NtWaitForMultipleObjects",
            "WaitForMultipleObjectsEx",
            "WaitForMultipleObjectsExImplementation",
            "RealMsgWaitForMultipleObjectsEx",
            "MsgWaitForMultipleObjects",
            "F_1152915508__________________________________",
            "F2166389______________________________________",
            "F_917831355___________________________________",
            "F1315696776________________________________",
            "F_1428703866________________________________",
        ]
        expected_proto_signature = " | ".join(expected_normalized_frames)

        assert result.extra["proto_signature"] == expected_proto_signature
        assert result.extra["normalized_frames"] == expected_normalized_frames
        assert result.notes == []

    def test_action_2_with_templates(self):
        sgr = rules.SignatureGenerationRule()

        crash_data = {
            "os": "Windows NT",
            "crashing_thread": 0,
            "threads": [FRAMES_FROM_JSON_DUMP_WITH_TEMPLATES],
        }
        result = generator.Result()

        # the call to be tested
        assert sgr.action(crash_data, result) is True

        assert result.signature == "Alpha<T>::Echo<T>"
        expected_normalized_frames = [
            "NtWaitForMultipleObjects",
            "Alpha<T>::Echo<T>",
            "WaitForMultipleObjectsExImplementation",
            "RealMsgWaitForMultipleObjectsEx",
            "MsgWaitForMultipleObjects",
            "F_1152915508__________________________________",
            "F2166389______________________________________",
            "F_917831355___________________________________",
            "F1315696776________________________________",
            "F_1428703866________________________________",
        ]
        expected_proto_signature = " | ".join(expected_normalized_frames)
        assert result.extra["proto_signature"] == expected_proto_signature
        assert result.extra["normalized_frames"] == expected_normalized_frames
        assert result.notes == []

    def test_action_2_with_templates_and_special_case(self):
        sgr = rules.SignatureGenerationRule()

        crash_data = {
            "os": "Windows NT",
            "crashing_thread": 0,
            "threads": [FRAMES_FROM_JSON_DUMP_WITH_TEMPLATES_AND_SPECIAL_CASE],
        }
        result = generator.Result()

        # the call to be tested
        assert sgr.action(crash_data, result) is True

        assert (
            result.signature
            == "<name omitted> | IPC::ParamTraits<mozilla::net::NetAddr>::Write"
        )
        expected_normalized_frames = [
            "NtWaitForMultipleObjects",
            "<name omitted>",
            "IPC::ParamTraits<mozilla::net::NetAddr>::Write",
            "RealMsgWaitForMultipleObjectsEx",
            "MsgWaitForMultipleObjects",
            "F_1152915508__________________________________",
            "F2166389______________________________________",
            "F_917831355___________________________________",
            "F1315696776________________________________",
            "F_1428703866________________________________",
        ]
        expected_proto_signature = " | ".join(expected_normalized_frames)
        assert result.extra["proto_signature"] == expected_proto_signature
        assert result.extra["normalized_frames"] == expected_normalized_frames
        assert result.notes == []

    def test_action_3(self):
        sgr = rules.SignatureGenerationRule()

        crash_data = {"thread": [[]]}
        result = generator.Result()

        # the call to be tested
        assert sgr.action(crash_data, result) is True

        assert result.signature == "EMPTY: no frame data available"
        assert "proto_signature" not in result.extra
        assert "normalized_frames" not in result.extra
        assert result.notes == [
            "SignatureGenerationRule: CSignatureTool: no frame data for crashing thread (0)"
        ]

    def test_lower_case_modules(self):
        sgr = rules.SignatureGenerationRule()

        crash_data = {
            "os": "Windows NT",
            "threads": [
                {
                    "frames": [
                        {"offset": "0x5e39bf21", "trust": "cfi"},
                        {"offset": "0x5e39bf21", "trust": "cfi"},
                        {"offset": "0x5e39bf21", "trust": "cfi"},
                        {
                            "frame": 3,
                            "module": "USER2.dll",
                            "module_offset": "0x20869",
                            "offset": "0x77370869",
                            "trust": "cfi",
                        },
                    ]
                }
            ],
        }
        result = generator.Result()

        # the call to be tested
        assert sgr.action(crash_data, result) is True
        assert result.signature == "user2.dll"
        expected_normalized_frames = [
            "@0x5e39bf21",
            "@0x5e39bf21",
            "@0x5e39bf21",
            "user2.dll@0x20869",
        ]
        expected_proto_signature = " | ".join(expected_normalized_frames)
        assert result.extra["proto_signature"] == expected_proto_signature
        assert result.extra["normalized_frames"] == expected_normalized_frames
        assert result.notes == []


class TestStackOverflowSignature:
    def test_predicate_no_match(self):
        result = generator.Result()
        result.signature = "hello"
        rule = rules.StackOverflowSignature()
        assert rule.predicate({}, result) is False

    @pytest.mark.parametrize(
        "reason, expected",
        [
            ("FOO", False),
            ("EXCEPTION_STACK_OVERFLOW", True),
        ],
    )
    def test_predicate_reason(self, reason, expected):
        crash_data = {"reason": reason}
        result = generator.Result()
        result.signature = "Text"
        rule = rules.StackOverflowSignature()
        assert rule.predicate(crash_data, result) is expected

    def test_action_success(self):
        crash_data = {"reason": "EXCEPTION_STACK_OVERFLOW"}
        result = generator.Result()
        result.signature = "hello"
        rule = rules.StackOverflowSignature()
        action_result = rule.action(crash_data, result)

        assert action_result is True
        assert result.signature == "stackoverflow | hello"


class TestOOMSignature:
    def test_predicate_no_match(self):
        result = generator.Result()
        result.signature = "hello"
        rule = rules.OOMSignature()
        assert rule.predicate({}, result) is False

    def test_predicate(self):
        crash_data = {"oom_allocation_size": 17}
        result = {"signature": "hello", "notes": []}
        rule = rules.OOMSignature()
        assert rule.predicate(crash_data, result) is True

    def test_predicate_signature_fragment_1(self):
        crash_data = {}
        result = generator.Result()
        result.signature = "this | is | a | NS_ABORT_OOM | signature"
        rule = rules.OOMSignature()
        assert rule.predicate(crash_data, result) is True

    def test_predicate_signature_fragment_2(self):
        crash_data = {}
        result = generator.Result()
        result.signature = "mozalloc_handle_oom | this | is | bad"
        rule = rules.OOMSignature()
        assert rule.predicate(crash_data, result) is True

    def test_predicate_signature_fragment_3(self):
        crash_data = {}
        result = generator.Result()
        result.signature = "CrashAtUnhandlableOOM"
        rule = rules.OOMSignature()
        assert rule.predicate(crash_data, result) is True

    @pytest.mark.parametrize(
        "crashdata, expected",
        [
            ({}, False),
            ({"crashing_thread": 5}, False),
            ({"crashing_thread": 5, "threads": []}, False),
            (
                {
                    "crashing_thread": 0,
                    "threads": [
                        {
                            "last_error_value": "FOO",
                        },
                    ],
                },
                False,
            ),
            (
                {
                    "crashing_thread": 0,
                    "threads": [
                        {
                            "last_error_value": "ERROR_COMMITMENT_LIMIT",
                        },
                    ],
                },
                True,
            ),
            (
                {
                    "crashing_thread": 0,
                    "threads": [
                        {
                            "last_error_value": "ERROR_NOT_ENOUGH_MEMORY",
                        },
                    ],
                },
                True,
            ),
        ],
    )
    def test_predicate_error_commitment_limit(self, crashdata, expected):
        result = generator.Result()
        result.signature = "Text"
        rule = rules.OOMSignature()
        assert rule.predicate(crashdata, result) is expected

    @pytest.mark.parametrize(
        "reason, expected",
        [
            ("FOO", False),
            ("STATUS_NO_MEMORY", True),
            ("STATUS_FATAL_MEMORY_EXHAUSTION", True),
        ],
    )
    def test_predicate_reason(self, reason, expected):
        crash_data = {"reason": reason}
        result = generator.Result()
        result.signature = "Text"
        rule = rules.OOMSignature()
        assert rule.predicate(crash_data, result) is expected

    @pytest.mark.parametrize(
        "value, expected",
        [
            (None, False),
            ("Reporting", True),
            ("Reported", False),
        ],
    )
    def test_predicate_js_large_allocation_failure(self, value, expected):
        crash_data = {"js_large_allocation_failure": value}
        result = generator.Result()
        result.signature = "Text"
        rule = rules.OOMSignature()
        assert rule.predicate(crash_data, result) is expected

    def test_action_success(self):
        crash_data = {}
        result = generator.Result()
        result.signature = "hello"
        rule = rules.OOMSignature()
        action_result = rule.action(crash_data, result)

        assert action_result is True
        assert result.signature == "OOM | unknown | hello"

    def test_action_small(self):
        crash_data = {"oom_allocation_size": 17}
        result = generator.Result()
        result.signature = "hello"
        rule = rules.OOMSignature()
        action_result = rule.action(crash_data, result)

        assert action_result is True
        assert result.signature == "OOM | small"

    def test_action_large(self):
        crash_data = {"oom_allocation_size": 17000000}
        result = generator.Result()
        result.signature = "hello"

        rule = rules.OOMSignature()
        action_result = rule.action(crash_data, result)

        assert action_result is True
        assert result.signature == "OOM | large | hello"

    def test_action_js_large_allocation_failure(self):
        crash_data = {"js_large_allocation_failure": "Reporting"}
        result = generator.Result()
        result.signature = "hello"

        rule = rules.OOMSignature()
        action_result = rule.action(crash_data, result)

        assert action_result is True
        assert result.signature == "OOM | large | hello"


class TestBadHardware:
    @pytest.mark.parametrize(
        "reason, expected",
        [
            ("FOO", False),
            ("EXCEPTION_IN_PAGE_ERROR_READ / STATUS_DEVICE_DATA_ERROR", True),
            ("EXCEPTION_IN_PAGE_ERROR_EXEC / STATUS_DEVICE_DATA_ERROR", True),
            ("EXCEPTION_IN_PAGE_ERROR_WRITE / STATUS_DEVICE_DATA_ERROR", True),
        ],
    )
    def test_predicate_reason(self, reason, expected):
        crash_data = {"reason": reason}
        result = generator.Result()
        result.signature = "Text"
        rule = rules.BadHardware()
        assert rule.predicate(crash_data, result) is expected

    def test_action(self):
        crash_data = {
            "reason": "EXCEPTION_IN_PAGE_ERROR_READ / STATUS_DEVICE_DATA_ERROR",
        }
        result = generator.Result()
        result.signature = "Text"
        rule = rules.BadHardware()
        action_result = rule.action(crash_data, result)
        assert action_result is True
        assert result.signature == "bad hardware | Text"


class TestAbortSignature:
    def test_predicate(self):
        rule = rules.AbortSignature()
        crash_data = {"abort_message": "something"}
        result = generator.Result()
        result.signature = "hello"
        assert rule.predicate(crash_data, result) is True

    def test_predicate_no_match(self):
        rule = rules.AbortSignature()
        # No AbortMessage
        crash_data = {}
        result = generator.Result()
        result.signature = "hello"
        assert rule.predicate(crash_data, result) is False

    def test_predicate_empty_message(self):
        rule = rules.AbortSignature()
        crash_data = {"abort_message": ""}
        result = generator.Result()
        result.signature = "hello"
        assert rule.predicate(crash_data, result) is False

    def test_action_success(self):
        rule = rules.AbortSignature()
        crash_data = {"abort_message": "unknown"}
        result = generator.Result()
        result.signature = "hello"
        action_result = rule.action(crash_data, result)
        assert action_result is True
        assert result.signature == "Abort | unknown | hello"

    def test_action_success_long_message(self):
        rule = rules.AbortSignature()

        crash_data = {"abort_message": "a" * 81}
        result = generator.Result()
        result.signature = "hello"
        action_result = rule.action(crash_data, result)

        assert action_result is True
        assert result.signature == "Abort | {}... | hello".format("a" * 77)

    @pytest.mark.parametrize(
        "abort_msg, expected",
        [
            # Test with just the "ABOR" thing at the start
            ("[5392] ###!!! ABORT: foo bar line 42", "Abort | foo bar line 42 | hello"),
            # Test with a file name and line number
            (
                (
                    "[7616] ###!!! ABORT: unsafe destruction: file "
                    "c:/builds/moz2_slave/m-rel-w32-00000000000000000000/build/src/"
                    "dom/plugins/ipc/PluginModuleParent.cpp, line 777"
                ),
                "Abort | unsafe destruction | hello",
            ),
            # Test with a message that lacks interesting content.
            ("[204] ###!!! ABORT: file ?, ", "Abort | hello"),
            # Test with another message that lacks interesting content.
            (
                (
                    "[4648] ###!!! ABORT: file resource:///modules/sessionstore/"
                    "SessionStore.jsm, line 1459"
                ),
                "Abort | hello",
            ),
            # Test with "unable to find a usable font" case
            (
                "unable to find a usable font (\u5fae\u8f6f\u96c5\u9ed1)",
                "Abort | unable to find a usable font | hello",
            ),
        ],
    )
    def test_action_success_remove_unwanted_parts(self, abort_msg, expected):
        rule = rules.AbortSignature()

        crash_data = {"abort_message": abort_msg}
        result = generator.Result()
        result.signature = "hello"
        action_result = rule.action(crash_data, result)

        assert action_result is True
        assert result.signature == expected

    def test_action_non_ascii_abort_message(self):
        # Non-ascii characters are removed from abort messages
        rule = rules.AbortSignature()
        crash_data = {"abort_message": "\u018a unknown"}
        result = generator.Result()
        result.signature = "hello"
        action_result = rule.action(crash_data, result)
        assert action_result is True
        assert result.signature == "Abort | unknown | hello"


class TestSigFixWhitespace:
    @pytest.mark.parametrize(
        "signature, expected",
        [
            # Leading and trailing whitespace are removed
            ("all   good", "all good"),
            ("all   good     ", "all good"),
            ("    all   good  ", "all good"),
            # Non-space whitespace is converted to spaces
            ("all\tgood", "all good"),
            ("all\n\ngood", "all good"),
            # Multiple consecutive spaces are converted to a single space
            ("all   good", "all good"),
            ("all  |  good", "all | good"),
        ],
    )
    def test_whitespace_fixing(self, signature, expected):
        rule = rules.SigFixWhitespace()
        result = generator.Result()
        result.signature = signature
        action_result = rule.action({}, result)
        assert action_result is True
        assert result.signature == expected


class TestSigTruncate:
    def test_predicate_no_match(self):
        rule = rules.SigTruncate()
        result = generator.Result()
        result.signature = "0" * 100
        assert rule.predicate({}, result) is False

    def test_predicate(self):
        rule = rules.SigTruncate()
        result = generator.Result()
        result.signature = "9" * 256
        assert rule.predicate({}, result) is True

    def test_action_success(self):
        rule = rules.SigTruncate()
        result = generator.Result()
        result.signature = "9" * 256
        action_result = rule.action({}, result)
        assert action_result is True
        assert len(result.signature) == 255
        assert result.signature.endswith("9...")


class TestStackwalkerErrorSignatureRule:
    def test_predicate_no_match_signature(self):
        rule = rules.StackwalkerErrorSignatureRule()
        result = generator.Result()
        result.signature = "0" * 100
        assert rule.predicate({}, result) is False

    def test_predicate_no_match_missing_mdsw_status_string(self):
        rule = rules.StackwalkerErrorSignatureRule()
        result = generator.Result()
        result.signature = "EMPTY: like my soul"
        assert rule.predicate({}, result) is False

    def test_predicate(self):
        rule = rules.StackwalkerErrorSignatureRule()
        crash_data = {"mdsw_status_string": "catastrophic stackwalker failure"}
        result = generator.Result()
        result.signature = "EMPTY: like my soul"
        assert rule.predicate(crash_data, result) is True

    def test_action_success(self):
        rule = rules.StackwalkerErrorSignatureRule()
        crash_data = {"mdsw_status_string": "catastrophic stackwalker failure"}
        result = generator.Result()
        result.signature = "EMPTY: like my soul"
        action_result = rule.action(crash_data, result)
        assert action_result is True
        expected = "EMPTY: like my soul; catastrophic stackwalker failure"
        assert result.signature == expected


class TestSignatureWatchDogRule:
    def test_predicate(self):
        srwd = rules.SignatureRunWatchDog()

        result = generator.Result()
        result.signature = "I'm not real"
        assert srwd.predicate({}, result) is False

        result = generator.Result()
        result.signature = "mozilla::`anonymous namespace''::RunWatchdog(void*)"
        assert srwd.predicate({}, result) is True

        result = generator.Result()
        result.signature = "mozilla::(anonymous namespace)::RunWatchdog"
        assert srwd.predicate({}, result) is True

    def test_action(self):
        sgr = rules.SignatureRunWatchDog()

        crash_data = {
            "os": "Windows NT",
            "crashing_thread": 0,
            "threads": [FRAMES_FROM_JSON_DUMP],
        }
        result = generator.Result()
        result.signature = "foo::bar"

        # the call to be tested
        assert sgr.action(crash_data, result) is True

        # Verify the signature has been re-generated based on thread 0.
        expected = (
            "shutdownhang | MsgWaitForMultipleObjects | "
            "F_1152915508__________________________________"
        )
        assert result.signature == expected
        assert result.notes == []


class TestSignatureIPCChannelError:
    def test_predicate_no_match(self):
        rule = rules.SignatureIPCChannelError()

        result = generator.Result()
        assert rule.predicate({}, result) is False

        crash_data = {"ipc_channel_error": ""}
        result = {"signature": "", "notes": []}
        assert rule.predicate(crash_data, result) is False

    def test_predicate(self):
        rule = rules.SignatureIPCChannelError()

        crash_data = {"ipc_channel_error": "foo, bar"}
        result = generator.Result()
        assert rule.predicate(crash_data, result) is True

    def test_regular_ipcchannelerror(self):
        rule = rules.SignatureIPCChannelError()

        channel_error = "ipc" * 50
        original_signature = "foo::bar"

        crash_data = {"ipc_channel_error": channel_error}
        result = generator.Result()
        result.signature = original_signature

        action_result = rule.action(crash_data, result)
        assert action_result is True
        expected = "IPCError-content | {}".format(channel_error[:100])
        assert result.signature == expected
        assert result.notes == [
            "SignatureIPCChannelError: IPC Channel Error stomped on signature"
        ]

        # Now test with a browser crash.
        crash_data["additional_minidumps"] = ["upload_file_minidump_browser"]
        result = generator.Result()
        result.signature = original_signature

        action_result = rule.action(crash_data, result)
        assert action_result is True

        assert result.signature == "IPCError-browser | {}".format(channel_error[:100])
        assert result.notes == [
            "SignatureIPCChannelError: IPC Channel Error stomped on signature"
        ]


class TestSignatureShutdownTimeout:
    def test_predicate_no_match(self):
        rule = rules.SignatureShutdownTimeout()
        result = generator.Result()
        assert rule.predicate({}, result) is False

    def test_predicate(self):
        rule = rules.SignatureShutdownTimeout()

        crash_data = {"async_shutdown_timeout": '{"foo": "bar"}'}
        result = generator.Result()
        assert rule.predicate(crash_data, result) is True

    def test_action_missing_valueerror(self):
        rule = rules.SignatureShutdownTimeout()

        crash_data = {
            # This will cause json.load to raise an error
            "async_shutdown_timeout": "{{{{"
        }
        result = generator.Result()
        result.signature = "foo"

        action_result = rule.action(crash_data, result)
        assert action_result is True
        assert result.signature == "AsyncShutdownTimeout | UNKNOWN"

        assert "Error parsing AsyncShutdownTimeout:" in result.notes[0]
        assert "Expecting property name" in result.notes[0]
        print(result.notes[1])
        assert (
            'Signature replaced with a Shutdown Timeout signature, was: "foo"'
            in result.notes[1]
        )

    def test_action_missing_keyerror(self):
        rule = rules.SignatureShutdownTimeout()

        crash_data = {
            "async_shutdown_timeout": json.dumps({"no": "phase or condition"})
        }
        result = generator.Result()
        result.signature = "foo"

        action_result = rule.action(crash_data, result)
        assert action_result is True
        assert result.signature == "AsyncShutdownTimeout | (unknown) | (none)"

        assert result.notes == [
            "SignatureShutdownTimeout: Signature replaced with a Shutdown "
            + 'Timeout signature, was: "foo"',
        ]

    def test_action_success(self):
        rule = rules.SignatureShutdownTimeout()

        crash_data = {
            "async_shutdown_timeout": json.dumps(
                {"phase": "beginning", "conditions": [{"name": "A"}, {"name": "B"}]}
            )
        }
        result = generator.Result()
        result.signature = "foo"

        action_result = rule.action(crash_data, result)
        assert action_result is True
        assert result.signature == "AsyncShutdownTimeout | beginning | A,B"
        assert result.notes == [
            "SignatureShutdownTimeout: Signature replaced with a Shutdown "
            + 'Timeout signature, was: "foo"',
        ]

    def test_action_success_string_conditions(self):
        rule = rules.SignatureShutdownTimeout()

        crash_data = {
            "async_shutdown_timeout": json.dumps(
                {"phase": "beginning", "conditions": ["A", "B", "C"]}
            )
        }
        result = generator.Result()
        result.signature = "foo"

        action_result = rule.action(crash_data, result)
        assert action_result is True
        assert result.signature == "AsyncShutdownTimeout | beginning | A,B,C"
        assert result.notes == [
            "SignatureShutdownTimeout: Signature replaced with a Shutdown "
            + 'Timeout signature, was: "foo"',
        ]

    def test_action_success_empty_conditions_key(self):
        rule = rules.SignatureShutdownTimeout()

        crash_data = {
            "async_shutdown_timeout": json.dumps(
                {"phase": "beginning", "conditions": []}
            )
        }
        result = generator.Result()
        result.signature = "foo"

        action_result = rule.action(crash_data, result)
        assert action_result is True
        assert result.signature == "AsyncShutdownTimeout | beginning | (none)"
        assert result.notes == [
            "SignatureShutdownTimeout: Signature replaced with a Shutdown "
            + 'Timeout signature, was: "foo"',
        ]

    def test_action_unsupported_structure(self):
        rule = rules.SignatureShutdownTimeout()

        crash_data = {"async_shutdown_timeout": json.dumps({})}
        result = generator.Result()
        result.signature = "foo"

        action_result = rule.action(crash_data, result)
        assert action_result is True
        assert result.signature == "AsyncShutdownTimeout | (unknown) | (none)"
        assert result.notes == [
            "SignatureShutdownTimeout: Signature replaced with a Shutdown "
            + 'Timeout signature, was: "foo"',
        ]


class TestSignatureIPCMessageName:
    def test_predicate_no_ipc_message_name(self):
        rule = rules.SignatureIPCMessageName()
        result = generator.Result()
        assert rule.predicate({}, result) is False

    def test_predicate_empty_string(self):
        rule = rules.SignatureIPCMessageName()
        crash_data = {"ipc_message_name": ""}
        result = generator.Result()
        assert rule.predicate(crash_data, result) is False

    def test_predicate(self):
        rule = rules.SignatureIPCMessageName()
        crash_data = {"ipc_message_name": "foo, bar"}
        result = generator.Result()
        result.signature = "fooo::baar"
        assert rule.predicate(crash_data, result) is True

    def test_action_success(self):
        rule = rules.SignatureIPCMessageName()
        crash_data = {"ipc_message_name": "foo, bar"}
        result = generator.Result()
        result.signature = "fooo::baar"
        action_result = rule.action(crash_data, result)
        assert action_result is True
        assert result.signature == "fooo::baar | IPC_Message_Name=foo, bar"


class TestSignatureParentIDNotEqualsChildID:
    def test_predicate_no_moz_crash_reason(self):
        rule = rules.SignatureParentIDNotEqualsChildID()
        result = {"signature": "", "notes": []}
        assert rule.predicate({}, result) is False

    def test_predicate_empty_moz_crash_reason(self):
        rule = rules.SignatureParentIDNotEqualsChildID()
        crash_data = {"moz_crash_reason": ""}
        result = {"signature": "", "notes": []}
        assert rule.predicate(crash_data, result) is False

    def test_predicate_match(self):
        rule = rules.SignatureParentIDNotEqualsChildID()
        crash_data = {
            "moz_crash_reason": "MOZ_RELEASE_ASSERT(parentBuildID == childBuildID)"
        }
        result = generator.Result()
        result.signature = "fooo::baar"
        assert rule.predicate(crash_data, result) is True

    def test_action(self):
        rule = rules.SignatureParentIDNotEqualsChildID()
        crash_data = {
            "moz_crash_reason": "MOZ_RELEASE_ASSERT(parentBuildID == childBuildID)"
        }
        result = generator.Result()
        result.signature = "fooo::baar"
        action_result = rule.action(crash_data, result)
        assert action_result is True
        assert result.signature == "parentBuildID != childBuildID"
        assert result.notes == [
            'SignatureParentIDNotEqualsChildID: Signature replaced with MOZ_RELEASE_ASSERT, was: "fooo::baar"'  # noqa
        ]


class TestHungProcess:
    def test_predicate_no_match(self):
        result = generator.Result()
        result.signature = "hello"
        rule = rules.HungProcess()
        assert rule.predicate({}, result) is False

    def test_predicate_match(self):
        crash_data = {"hang": "ui"}
        result = generator.Result()
        result.signature = "Text"
        rule = rules.HungProcess()
        assert rule.predicate(crash_data, result) is True

    def test_action_success(self):
        crash_data = {"hang": "ui"}
        result = generator.Result()
        result.signature = "Text"
        rule = rules.HungProcess()
        action_result = rule.action(crash_data, result)

        assert action_result is True
        assert result.signature == "hang: ui | Text"



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.


# x.y.YYYYMMDD or x.y.YYYYMMDD.dev0 -- semver
__version__ = "2.0.20231009"



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import dataclasses
import inspect
import sys
from typing import Any, Dict, List

from .rules import (
    AbortSignature,
    BadHardware,
    HungProcess,
    OOMSignature,
    SigFixWhitespace,
    SignatureGenerationRule,
    SignatureIPCChannelError,
    SignatureIPCMessageName,
    SignatureParentIDNotEqualsChildID,
    SignatureRunWatchDog,
    SignatureShutdownTimeout,
    SigTruncate,
    StackOverflowSignature,
    StackwalkerErrorSignatureRule,
)


DEFAULT_RULESET = [
    SignatureGenerationRule,
    StackwalkerErrorSignatureRule,
    BadHardware,
    OOMSignature,
    AbortSignature,
    SignatureShutdownTimeout,
    SignatureRunWatchDog,
    SignatureIPCChannelError,
    SignatureIPCMessageName,
    SignatureParentIDNotEqualsChildID,
    StackOverflowSignature,
    HungProcess,
    # NOTE(willkg): These should always come last and in this order
    SigFixWhitespace,
    SigTruncate,
]


@dataclasses.dataclass
class Result:
    signature: str = ""
    notes: List[str] = dataclasses.field(default_factory=list, repr=False)
    debug_log: List[str] = dataclasses.field(default_factory=list, repr=False)
    extra: Dict[str, Any] = dataclasses.field(default_factory=dict, repr=False)

    def set_signature(self, rule, signature):
        self.debug(rule, 'change signature: "%s" -> "%s"', self.signature, signature)
        self.signature = signature

    def info(self, rule, msg, *args):
        if args:
            msg = msg % args
        self.notes.append("%s: %s" % (rule, msg))

    def debug(self, rule, msg, *args):
        if args:
            msg = msg % args
        self.debug_log.append("%s: %s" % (rule, msg))

    def to_dict(self):
        return {
            "signature": self.signature,
            "notes": self.notes,
            "debug_log": self.debug_log,
            "extra": self.extra,
        }


class SignatureGenerator:
    def __init__(self, ruleset=None, error_handler=None, **cfg):
        """
        :param ruleset: list of rule classes to use for signature generation
        :param error_handler: error handling function with signature
            ``fun(signature_data, exc_info, extra)``
        :param cfg: configuration keyword arguments to pass to rules being
            instantiated

        """
        self.error_handler = error_handler
        self.pipeline = self.initialize_pipeline(
            ruleset or list(DEFAULT_RULESET), **cfg
        )

    def initialize_pipeline(self, ruleset, **cfg):
        """Initialize a pipeline of rules

        For each rule in the ruleset:

        * if the rule is an instance, adds it to pipeline
        * if the rule is a class, uses introspection to determine the configuration
          required and instantiates the rule class with that

        :param ruleset: list of rule instances and classes
        :param cfg: configuration kwargs to pass to rules being instantiated; see each
            rule's documentation for what configuration is supported

        :returns: list of rule instances

        """
        pipeline = []
        for rule in ruleset:
            instance = None
            if inspect.isclass(rule):
                fun_signature = inspect.signature(rule)
                kwargs = {}
                for key, param in fun_signature.parameters.items():
                    if (
                        param.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD
                        and key in cfg
                    ):
                        kwargs[key] = cfg[key]

                instance = rule(**kwargs)
            else:
                instance = rule
            pipeline.append(instance)
        return pipeline

    def generate(self, signature_data):
        """Takes data and returns a signature

        :arg dict signature_data: data to use to generate a signature

        :returns: ``Result`` instance

        """
        result = Result()

        for rule in self.pipeline:
            rule_name = rule.__class__.__name__

            try:
                if rule.predicate(signature_data, result):
                    rule.action(signature_data, result)

            except Exception as exc:
                if self.error_handler:
                    self.error_handler(
                        signature_data,
                        exc_info=sys.exc_info(),
                        extra={"rule": rule_name},
                    )
                result.info(rule_name, "Rule failed: %s", exc)

        return result



#!/usr/bin/env python

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import codecs
import os
import re

from setuptools import setup, find_packages


def get_long_desc():
    return codecs.open("README.rst", encoding="utf-8").read()


def get_version():
    fn = os.path.join("siggen", "__init__.py")
    vsre = r"""^__version__ = ['"]([^'"]*)['"]"""
    version_file = codecs.open(fn, mode="r", encoding="utf-8").read()
    return re.search(vsre, version_file, re.M).group(1)


INSTALL_REQUIRES = [
    "glom",
    "importlib_resources",
]
EXTRAS_REQUIRE = {
    "cli": [
        "requests<3",
    ],
}


setup(
    name="siggen",
    version=get_version(),
    description="Experimental extraction of Socorro signature generation",
    long_description=get_long_desc(),
    long_description_content_type="text/x-rst",
    maintainer="Will Kahn-Greene",
    maintainer_email="willkg@mozilla.com",
    url="https://github.com/willkg/socorro-siggen",
    license="Mozilla Public License v2",
    packages=find_packages(),
    include_package_data=True,
    install_requires=INSTALL_REQUIRES,
    extras_require=EXTRAS_REQUIRE,
    python_requires=">=3.8",
    entry_points="""
        [console_scripts]
        signify=siggen.cmd_signify:main
        fetch-data=siggen.cmd_fetch_data:main
        signature=siggen.cmd_signature:main
    """,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: Implementation :: CPython",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)
